# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import types

from six import moves
import testscenarios

from subunit2sql.db import api
from subunit2sql.db import models
from subunit2sql.tests import base
from subunit2sql.tests import db_test_utils
from subunit2sql.tests import subunit2sql_fixtures as fixtures

load_tests = testscenarios.load_tests_apply_scenarios


class TestDatabaseAPI(base.TestCase):

    scenarios = [
        ('mysql', {'dialect': 'mysql'}),
        ('postgresql', {'dialect': 'postgres'}),
        ('sqlite', {'dialect': 'sqlite'})
    ]

    def setUp(self):
        super(TestDatabaseAPI, self).setUp()
        self.useFixture(fixtures.LockFixture(self.dialect))
        if not db_test_utils.is_backend_avail(self.dialect):
            raise self.skipTest('%s is not available' % self.dialect)
        if self.dialect == 'mysql':
            self.useFixture(fixtures.MySQLConfFixture())
        elif self.dialect == 'postgres':
            self.useFixture(fixtures.PostgresConfFixture())
        elif self.dialect == 'sqlite':
            self.useFixture(fixtures.SqliteConfFixture())
        self.useFixture(fixtures.Database())

    def test_create_test(self):
        api.create_test('1234')
        res = api.get_all_tests()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].test_id, '1234')

    def test_create_test_and_get_by_test_id(self):
        create_res = api.create_test('fake_test', 2, 1, 1, 1.2)
        res = api.get_test_by_test_id('fake_test')
        self.assertEqual(res.id, create_res.id)
        self.assertEqual(res.test_id, 'fake_test')
        self.assertEqual(res.run_time, 1.2)
        self.assertEqual(res.run_count, 2)

    def test_get_test_by_test_id_invalid_id(self):
        res = api.get_test_by_test_id('fake_test')
        self.assertIsNone(res)

    def test_get_tests_by_test_ids(self):
        test_a = api.create_test('fake_test1', 2, 1, 1, 1.2)
        test_b = api.create_test('fake_test2', 4, 2, 2, 2.3)
        test_c = api.create_test('fake_test3', 6, 3, 3, 3.3)
        test_d = api.create_test('fake_test4', 8, 4, 4, 4.3)
        result = api.get_tests_by_test_ids(
            ['fake_test1', 'fake_test2', 'fake_test3'])
        result_ids = [x.id for x in result]
        self.assertIn(test_a.id, result_ids)
        self.assertIn(test_b.id, result_ids)
        self.assertIn(test_c.id, result_ids)
        self.assertNotIn(test_d.id, result_ids)

    def test_get_tests_by_test_ids_no_matches(self):
        api.create_test('fake_test5', 2, 1, 1, 1.2)
        api.create_test('fake_test6', 4, 2, 2, 2.3)
        api.create_test('fake_test7', 6, 3, 3, 3.3)
        api.create_test('fake_test8', 8, 4, 4, 4.3)
        result = api.get_tests_by_test_ids(
            ['fake_test1', 'fake_test2', 'fake_test3'])
        self.assertEqual([], result)

    def test_get_test_with_sql_injection(self):
        api.create_test("test_terror'); DROP TABLE tests;")
        res = api.get_all_tests()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].test_id, "test_terror'); DROP TABLE tests;")

    def test_create_run_and_list(self):
        res = api.create_run()
        self.assertIsNotNone(res)
        all_runs = api.get_all_runs()
        self.assertEqual(len(all_runs), 1)
        self.assertEqual(res.id, all_runs[0].id)

    def test_update_run_and_list(self):
        res = api.create_run()
        dt = datetime.datetime.utcnow()
        # NOTE(masayukig): The DateTime column in some DB envs can't store a
        # microseconds resolution value. So this removes the microseconds value
        # here.
        dt = dt.replace(microsecond=0)
        dt_dummy = dt - datetime.timedelta(days=1)
        res_dummy = api.create_run(run_at=dt_dummy)

        values = {'skips': 98, 'fails': 97, 'passes': 96, 'run_time': 1.123,
                  'artifacts': 'fake_url', 'run_at': dt}
        updated_res = api.update_run(values, res.id)
        all_runs = api.get_all_runs()
        self.assertEqual(len(all_runs), 2)
        for run in all_runs:
            if run.id == res.id:
                self.assertEqual(updated_res.id, run.id)
                for key in values:
                    self.assertEqual(values[key], run[key])
                    self.assertEqual(run[key], updated_res[key])
            elif run.id == res_dummy.id:
                self.assertNotEqual(updated_res.id, run.id)
                for key in values:
                    if key == 'run_at':
                        continue
                    self.assertNotEqual(values[key], run[key])
                    self.assertNotEqual(run[key], updated_res[key])
            else:
                self.fail('an unexpected run(%s) was found' % run.id)

    def test_get_test_runs_dicts_with_no_meta(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        start_time = datetime.datetime.utcnow().replace(microsecond=0)
        stop_time = datetime.datetime.utcnow().replace(microsecond=0)
        api.create_test_run(test_a.id, run.id, 'success',
                            start_time, stop_time)
        test_run_dict = api.get_tests_run_dicts_from_run_id(run.uuid)
        self.assertEqual(1, len(test_run_dict))
        self.assertIn('fake_test', test_run_dict)
        self.assertEqual(test_run_dict['fake_test']['status'], 'success')
        self.assertEqual(test_run_dict['fake_test']['start_time'], start_time)
        self.assertEqual(test_run_dict['fake_test']['stop_time'], stop_time)
        self.assertNotIn('metadata', test_run_dict['fake_test'])

    def test_get_test_runs_dicts_with_no_stop_time(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        start_time = datetime.datetime.utcnow().replace(microsecond=0)
        stop_time = None
        api.create_test_run(test_a.id, run.id, 'success',
                            start_time, stop_time)
        test_run_dict = api.get_tests_run_dicts_from_run_id(run.uuid)
        self.assertEqual(1, len(test_run_dict))
        self.assertIn('fake_test', test_run_dict)
        self.assertEqual(test_run_dict['fake_test']['status'], 'success')
        self.assertEqual(test_run_dict['fake_test']['start_time'], start_time)
        self.assertEqual(test_run_dict['fake_test']['stop_time'], stop_time)

    def test_get_test_runs_dicts_with_no_start_time(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        stop_time = datetime.datetime.utcnow().replace(microsecond=0)
        start_time = None
        api.create_test_run(test_a.id, run.id, 'success',
                            start_time, stop_time)
        test_run_dict = api.get_tests_run_dicts_from_run_id(run.uuid)
        self.assertEqual(1, len(test_run_dict))
        self.assertIn('fake_test', test_run_dict)
        self.assertEqual(test_run_dict['fake_test']['status'], 'success')
        self.assertEqual(test_run_dict['fake_test']['start_time'], start_time)
        self.assertEqual(test_run_dict['fake_test']['stop_time'], stop_time)

    def test_get_test_runs_dicts_with_no_start_or_stop_time(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        stop_time = None
        start_time = None
        api.create_test_run(test_a.id, run.id, 'success',
                            start_time, stop_time)
        test_run_dict = api.get_tests_run_dicts_from_run_id(run.uuid)
        self.assertEqual(1, len(test_run_dict))
        self.assertIn('fake_test', test_run_dict)
        self.assertEqual(test_run_dict['fake_test']['status'], 'success')
        self.assertEqual(test_run_dict['fake_test']['start_time'], start_time)
        self.assertEqual(test_run_dict['fake_test']['stop_time'], stop_time)

    def test_get_test_runs_dicts_with_meta(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        test_run = api.create_test_run(test_a.id, run.id, 'success',
                                       datetime.datetime.utcnow(),
                                       datetime.datetime.utcnow())
        run_meta = {
            'key_a': 'value_b',
            'key_b': 'value_a',
            'attrs': 'test,smoke,notatest',
        }
        api.add_test_run_metadata(run_meta, test_run.id)
        test_run_dict = api.get_tests_run_dicts_from_run_id(run.uuid)
        self.assertEqual(3, len(test_run_dict['fake_test']['metadata']))
        for meta in run_meta:
            self.assertIn(meta, test_run_dict['fake_test']['metadata'])
            self.assertEqual(run_meta[meta],
                             test_run_dict['fake_test']['metadata'][meta])

    def test_create_test_run_and_list(self):
        run = api.create_run()
        test = api.create_test('fake_test')
        test_run = api.create_test_run(test.id, run.id, 'fail')
        self.assertIsNotNone(test_run)
        all_test_runs = api.get_all_test_runs()
        self.assertEqual(len(all_test_runs), 1)
        self.assertEqual(test_run.id, all_test_runs[0].id)

    def test_get_test_runs_dicts_from_run_id_are_in_chrono_order(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        test_b = api.create_test('fake_test_2')
        test_c = api.create_test('fake_test_3')
        api.create_test_run(test_a.id, run.id, 'success',
                            datetime.datetime.utcnow())
        api.create_test_run(test_b.id, run.id, 'success',
                            datetime.datetime(1914, 6, 28, 10, 45, 0))
        api.create_test_run(test_c.id, run.id, 'success',
                            datetime.datetime(2014, 8, 26, 20, 00, 00))
        test_run_dicts = api.get_tests_run_dicts_from_run_id(run.uuid)
        self.assertEqual(len(test_run_dicts), 3)
        prev = None
        for test_run in test_run_dicts:
            if prev == None:
                prev = test_run
                continue
            self.assertTrue(test_run_dicts[test_run]['start_time'] >
                            test_run_dicts[prev]['start_time'])
            prev = test_run

    def test_get_test_run_duration(self):
        start_time = datetime.datetime.utcnow()
        stop_time = start_time + datetime.timedelta(minutes=3)
        run = api.create_run()
        test_a = api.create_test('fake_test')
        test_run = api.create_test_run(test_a.id, run.id, 'success',
                                       start_time, stop_time)
        dur = api.get_test_run_duration(test_run.id)
        self.assertEqual(180.0, dur)

    def test_get_id_from_test_id(self):
        test_a = api.create_test('fake_test')
        id_value = api.get_id_from_test_id('fake_test')
        self.assertEqual(test_a.id, id_value)

    def test_get_run_times_all_test_runs(self):
        timestamp_a = datetime.datetime.utcnow()
        timestamp_b = timestamp_a + datetime.timedelta(seconds=3)
        run_a = api.create_run()
        run_b = api.create_run()
        test_a = api.create_test('test_a')
        test_b = api.create_test('test_b')
        api.create_test_run(test_a.id, run_a.id, 'success', timestamp_a,
                            timestamp_b)
        api.create_test_run(test_a.id, run_b.id, 'success', timestamp_a,
                            timestamp_b)
        api.create_test_run(test_b.id, run_b.id, 'success', timestamp_a,
                            timestamp_b)
        res = api.get_run_times_all_test_runs()
        expected_dict = {
            'test_a': [3, 3],
            'test_b': [3]
        }
        self.assertEqual(expected_dict, res)

    def test_get_run_times_all_test_runs_with_tests_filter(self):
        timestamp_a = datetime.datetime.utcnow()
        timestamp_b = timestamp_a + datetime.timedelta(seconds=3)
        run_a = api.create_run()
        run_b = api.create_run()
        test_a = api.create_test('test_a')
        test_b = api.create_test('test_b')
        api.create_test_run(test_a.id, run_a.id, 'success', timestamp_a,
                            timestamp_b)
        api.create_test_run(test_a.id, run_b.id, 'success', timestamp_a,
                            timestamp_b)
        api.create_test_run(test_b.id, run_b.id, 'success', timestamp_a,
                            timestamp_b)
        res = api.get_run_times_all_test_runs(tests=['test_a'])
        expected_dict = {
            'test_a': [3, 3],
        }
        self.assertEqual(expected_dict, res)

    def test_get_test_runs_by_run_id(self):
        run_b = api.create_run()
        run_a = api.create_run()
        run_c = api.create_run()
        test_a = api.create_test('fake_test')
        testrun_a = api.create_test_run(test_a.id, run_a.id, 'success',
                                        datetime.datetime.utcnow())
        testrun_b = api.create_test_run(test_a.id, run_b.id, 'success',
                                        datetime.datetime.utcnow())
        testrun_c = api.create_test_run(test_a.id, run_c.id, 'success',
                                        datetime.datetime.utcnow())
        test_runs_a = api.get_test_runs_by_run_id(run_a.uuid)
        test_runs_b = api.get_test_runs_by_run_id(run_b.uuid)
        test_runs_c = api.get_test_runs_by_run_id(run_c.uuid)
        self.assertEqual(len(test_runs_a), 1)
        self.assertEqual(testrun_a.id, test_runs_a[0].id)
        self.assertEqual(testrun_a.status, test_runs_a[0].status)
        self.assertEqual(len(test_runs_b), 1)
        self.assertEqual(testrun_b.id, test_runs_b[0].id)
        self.assertEqual(testrun_b.status, test_runs_b[0].status)
        self.assertEqual(len(test_runs_c), 1)
        self.assertEqual(testrun_c.id, test_runs_c[0].id)
        self.assertEqual(testrun_c.status, test_runs_c[0].status)

    def test_get_runs_by_status_grouped_by_run_metadata(self):
        # Generating 20 runs:
        # 10 with no failures
        # 10 with 10 failures
        # 7 in 2010/2011 each, 6 in 2012
        # 10 in projecta/projectb each
        for i in moves.range(20):
            if i % 2 == 1:
                fails = 10
            else:
                fails = 0
            year = 2010 + (i % 3)
            run_at = datetime.datetime(year, 1, i + 1, 12, 0, 0)
            run = api.create_run(fails=fails, passes=10, run_at=run_at)
            self.assertIsNotNone(run)
            if i < 10:
                project = 'projecta'
            else:
                project = 'projectb'
            meta_dict = {'project': project}
            api.add_run_metadata(meta_dict, run.id)
        result = api.get_runs_by_status_grouped_by_run_metadata(
            'project', start_date='2012-01-01', stop_date='2012-12-31')
        # There should be two projects
        self.assertEqual(2, len(result.keys()))
        self.assertTrue('projecta' in result)
        self.assertTrue('projectb' in result)
        # There should be passes and failures
        self.assertEqual(2, len(result['projecta'].keys()))
        self.assertTrue('pass' in result['projecta'])
        self.assertTrue('fail' in result['projecta'])
        self.assertEqual(2, len(result['projectb'].keys()))
        self.assertTrue('pass' in result['projectb'])
        self.assertTrue('fail' in result['projectb'])

        self.assertEqual(2, result['projecta']['pass'])
        self.assertEqual(1, result['projecta']['fail'])
        self.assertEqual(1, result['projectb']['pass'])
        self.assertEqual(2, result['projectb']['fail'])

    def test_get_time_series_runs_by_key_value(self):
        runs = []
        run_at = datetime.datetime.utcnow()
        for run_num in moves.range(15):
            run = api.create_run(run_num, run_num + 1, run_num + 2, 3,
                                 run_at=run_at)
            runs.append(run)
            run_meta = {'test_key': 'fun', 'non_test': 'value-%s' % run_num}
            if run_num >= 3:
                run_meta = {'test_key': 'no-fun',
                            'non_test': 'value-%s' % run_num}
            api.add_run_metadata(run_meta, run.id)
        runs_time_series = api.get_time_series_runs_by_key_value('test_key',
                                                                 'fun')
        self.assertEqual(1, len(runs_time_series))
        timestamp = list(runs_time_series.keys())[0]
        self.assertEqual(3, len(runs_time_series[timestamp]))
        for run_num in moves.range(3):
            run_dict = {
                'skip': run_num,
                'fail': run_num + 1,
                'pass': run_num + 2,
                'id': runs[run_num].uuid,
                'run_time': 3.0,
                'metadata': {
                    u'test_key': u'fun',
                    u'non_test': u'value-%s' % run_num
                }
            }
            self.assertIn(run_dict, runs_time_series[timestamp])
        for run_num in moves.range(3, 14):
            missing_run_dict = {
                'skip': run_num,
                'fail': run_num + 1,
                'pass': run_num + 2,
                'id': runs[run_num].id,
                'run_time': 3.0,
                'metadata': {
                    u'test_key': u'fun',
                    u'non_test': u'value-%s' % run_num
                }
            }
            self.assertNotIn(missing_run_dict, runs_time_series[timestamp])

    def test_get_test_runs_test_test_id(self):
        run = api.create_run()
        test_a = api.create_test('fake_test')
        test_b = api.create_test('less_fake_test')
        api.create_test_run(test_a.id, run.id, 'success')
        api.create_test_run(test_b.id, run.id, 'success')
        res = api.get_test_runs_by_test_test_id('less_fake_test')
        self.assertEqual(1, len(res))
        self.assertEqual(test_b.id, res[0].test_id)
        self.assertEqual(run.id, res[0].run_id)

    def test_get_test_runs_test_test_id_with_run_metadata(self):
        run_a = api.create_run()
        run_b = api.create_run()
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'b_key': 'b_value'}, run_b.id)
        test_a = api.create_test('fake_test')
        test_b = api.create_test('less_fake_test')
        api.create_test_run(test_a.id, run_a.id, 'success')
        api.create_test_run(test_a.id, run_b.id, 'success')
        api.create_test_run(test_b.id, run_a.id, 'success')
        api.create_test_run(test_b.id, run_b.id, 'success')
        res = api.get_test_runs_by_test_test_id('less_fake_test', key='a_key',
                                                value='a_value')
        self.assertEqual(1, len(res))
        self.assertEqual(test_b.id, res[0].test_id)
        self.assertEqual(run_a.id, res[0].run_id)

    def test_get_all_run_metadata_keys(self):
        run = api.create_run()
        meta_dict = {
            'test_a': 'a',
            'test_a': 'b',
            'test_b': 'a',
            'test_c': 'a',
            'test_d': 'a',
            'test_c': 'b',
        }
        api.add_run_metadata(meta_dict, run.id)
        keys = api.get_all_run_metadata_keys()
        self.assertEqual(sorted(['test_a', 'test_b', 'test_c', 'test_d']),
                         sorted(keys))

    def test_get_all_test_metadata_keys(self):
        test = api.create_test('fake_test')
        meta_dict = {
            'test_a': 'a',
            'test_a': 'b',
            'test_b': 'a',
            'test_c': 'a',
            'test_d': 'a',
            'test_c': 'b',
        }
        api.add_test_metadata(meta_dict, test.id)
        keys = api.get_all_test_metadata_keys()
        self.assertEqual(sorted(['test_a', 'test_b', 'test_c', 'test_d']),
                         sorted(keys))

    def test_get_all_test_run_metadata_keys(self):
        run = api.create_run()
        test = api.create_test('fake_test')
        test_run = api.create_test_run(test.id, run.id, 'skip')
        meta_dict = {
            'test_a': 'a',
            'test_a': 'b',
            'test_b': 'a',
            'test_c': 'a',
            'test_d': 'a',
            'test_c': 'b',
        }
        api.add_test_run_metadata(meta_dict, test_run.id)
        keys = api.get_all_test_run_metadata_keys()
        self.assertEqual(sorted(['test_a', 'test_b', 'test_c', 'test_d']),
                         sorted(keys))

    def test_get_test_run_series(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=2)
        api.create_run(passes=5, run_at=timestamp_a)
        api.create_run(fails=2, run_at=timestamp_b)
        result = api.get_test_run_series(key=None, value=None)
        self.assertEqual(2, len(result.keys()))
        self.assertIn(timestamp_a, result.keys())
        self.assertIn(timestamp_b, result.keys())
        for timestamp in result:
            if timestamp.replace(
                    microsecond=0) == timestamp_a.replace(microsecond=0):
                self.assertEqual(5, result[timestamp])
            else:
                self.assertEqual(2, result[timestamp])

    def test_get_test_run_series_with_meta(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=2)
        run_a = api.create_run(passes=5, run_at=timestamp_a)
        api.create_run(fails=2, run_at=timestamp_b)
        api.add_run_metadata({'not_a_key': 'not_a_value'}, run_a.id)
        result = api.get_test_run_series(key='not_a_key',
                                         value='not_a_value')
        self.assertEqual(1, len(result.keys()))
        self.assertIn(timestamp_a, result.keys())
        self.assertNotIn(timestamp_b, result.keys())
        self.assertEqual(5, result[list(result.keys())[0]])

    def test_get_run_times_grouped_by_run_metadata_key(self):
        run_a = api.create_run(run_time=2.2, passes=2)
        run_b = api.create_run(run_time=3.5, passes=3)
        api.add_run_metadata({'key': 'value_a'}, run_a.id)
        api.add_run_metadata({'key': 'value_b'}, run_b.id)
        res = api.get_run_times_grouped_by_run_metadata_key('key')
        expected_res = {'value_a': [2.2], 'value_b': [3.5]}
        self.assertEqual(expected_res, res)

    def test_get_run_times_grouped_by_run_metadata_key_with_match_filter(self):
        run_a = api.create_run(run_time=2.2, passes=2)
        run_b = api.create_run(run_time=3.5, passes=3)
        run_c = api.create_run(run_time=75.432, passes=11)
        api.add_run_metadata({'key': 'value_a'}, run_a.id)
        api.add_run_metadata({'match_key': 'value_c'}, run_a.id)
        api.add_run_metadata({'key': 'value_b'}, run_b.id)
        api.add_run_metadata({'match_key': 'value_c'}, run_b.id)
        api.add_run_metadata({'key': 'value_b'}, run_c.id)
        res = api.get_run_times_grouped_by_run_metadata_key(
            'key', match_key='match_key', match_value='value_c')
        expected_res = {'value_a': [2.2], 'value_b': [3.5]}
        self.assertEqual(expected_res, res)

    def test_get_run_time_series_grouped_by_run_metadata_key(self):
        run_at_a = datetime.datetime(1914, 11, 11, 11, 11, 11)
        run_a = api.create_run(run_time=2.2, passes=2, run_at=run_at_a)
        run_at_b = datetime.datetime.utcnow().replace(microsecond=0)
        run_b = api.create_run(run_time=3.5, passes=3, run_at=run_at_b)
        api.add_run_metadata({'key': 'value_a'}, run_a.id)
        api.add_run_metadata({'key': 'value_b'}, run_b.id)
        res = api.get_run_times_time_series_grouped_by_run_metadata_key('key')
        expected_res = {run_at_a: {'value_a': 2.2},
                        run_at_b: {'value_b': 3.5}}
        self.assertEqual(expected_res, res)

    def test_get_run_ts_grouped_by_run_meta_key_with_match_filter(self):
        run_at_a = datetime.datetime(1914, 11, 11, 11, 11, 11)
        run_a = api.create_run(run_time=2.2, passes=2, run_at=run_at_a)
        run_at_b = datetime.datetime.utcnow().replace(microsecond=0)
        run_b = api.create_run(run_time=3.5, passes=3, run_at=run_at_b)
        run_c = api.create_run(run_time=75.432, passes=11)
        api.add_run_metadata({'key': 'value_a'}, run_a.id)
        api.add_run_metadata({'match_key': 'value_c'}, run_a.id)
        api.add_run_metadata({'key': 'value_b'}, run_b.id)
        api.add_run_metadata({'match_key': 'value_c'}, run_b.id)
        api.add_run_metadata({'key': 'value_b'}, run_c.id)
        res = api.get_run_times_time_series_grouped_by_run_metadata_key(
            'key', match_key='match_key', match_value='value_c')
        expected_res = {run_at_a: {'value_a': 2.2},
                        run_at_b: {'value_b': 3.5}}
        self.assertEqual(expected_res, res)

    def test_get_test_run_dict_by_run_meta_key_value(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=2)
        run_a = api.create_run()
        run_b = api.create_run()
        api.add_run_metadata({'key': 'true'}, run_a.id)
        api.add_run_metadata({'key': 'not so true'}, run_b.id)
        test_a = api.create_test('fake_test')
        api.create_test_run(test_a.id, run_a.id, 'success', timestamp_a,
                            timestamp_b)
        api.create_test_run(test_a.id, run_b.id, 'fail', timestamp_a,
                            datetime.datetime.utcnow().replace(microsecond=0))
        test_run_dicts = api.get_test_run_dict_by_run_meta_key_value('key',
                                                                     'true')
        self.assertEqual(1, len(test_run_dicts))
        self.assertEqual([{
            'test_id': 'fake_test',
            'status': 'success',
            'start_time': timestamp_a,
            'stop_time': timestamp_b,
        }], test_run_dicts)

    def test_delete_old_runs(self):
        run_a = api.create_run(run_at=datetime.datetime(
            1914, 6, 28, 10, 45, 0))
        run_b = api.create_run()
        api.add_run_metadata({'key': 'value'}, run_b.id)
        api.add_run_metadata({'key': 'not_so_much_a_value'}, run_a.id)
        api.delete_old_runs()
        runs = api.get_all_runs()
        self.assertEqual(1, len(runs))
        self.assertEqual(1, api.get_session().query(
            models.RunMetadata.id).count())
        self.assertEqual(run_b.id, runs[0].id)
        self.assertEqual(1, len(api.get_run_metadata(run_b.uuid)))
        self.assertEqual(0, len(api.get_run_metadata(run_a.uuid)))

    def test_delete_old_test_runs(self):
        run_a = api.create_run()
        run_b = api.create_run()
        test = api.create_test('fake_test')
        test_run_a = api.create_test_run(test.id, run_a.id, 'fail',
                                         start_time=datetime.datetime(
                                             1914, 6, 28, 10, 45, 0))
        test_run_b = api.create_test_run(test.id, run_b.id, 'fail',
                                         start_time=datetime.datetime.utcnow())
        api.add_test_run_metadata({'key': 'value'}, test_run_b.id)
        api.add_test_run_metadata({'key': 'not_so_much_a_value'},
                                  test_run_a.id)
        api.delete_old_test_runs()
        test_runs = api.get_all_test_runs()
        self.assertEqual(1, len(test_runs))
        self.assertEqual(test_run_b.id, test_runs[0].id)
        self.assertEqual(1, len(api.get_test_run_metadata(test_run_b.id)))
        self.assertEqual(0, len(api.get_test_run_metadata(test_run_a.id)))

    def test_delete_run_by_uuid(self):
        metadata_a = {'key': 'This is the metadata we\'re keeping'}
        metadata_b = {'key': 'This metadata should be removed'}
        run_a = api.create_run()
        run_b = api.create_run()
        uuid_to_delete = run_a.uuid
        api.add_run_metadata(metadata_a, run_a.id)
        api.add_run_metadata(metadata_b, run_b.id)
        api.delete_run_by_uuid(uuid_to_delete)
        runs = api.get_all_runs()
        self.assertEqual(1, len(runs))
        self.assertEqual(1, api.get_session().query(
            models.RunMetadata.id).count())
        self.assertEqual(run_b.id, runs[0].id)
        self.assertEqual(1, len(api.get_run_metadata(run_b.uuid)))
        self.assertEqual(0, len(api.get_run_metadata(run_a.uuid)))

    def test_delete_test_runs_by_run_uuid(self):
        # create runs
        run_a = api.create_run()
        run_b = api.create_run()
        uuid_to_delete = run_a.uuid
        # create test runs
        test = api.create_test('fake_test')
        test_run_a = api.create_test_run(test.id, run_a.id, 'fail',
                                         start_time=datetime.datetime(
                                             1914, 6, 28, 10, 45, 0))
        test_run_b = api.create_test_run(test.id, run_b.id, 'fail',
                                         start_time=datetime.datetime.utcnow())
        # initialize and metadata to test_runs
        metadata_a = {
            'key': 'this test belongs to the run we are deleting from'}
        metadata_b = {
            'key': 'this test belongs to the run we are leaving alone'}
        api.add_test_run_metadata(metadata_a, test_run_a.id)
        api.add_test_run_metadata(metadata_b, test_run_b.id)
        api.delete_test_runs_by_run_uuid(uuid_to_delete)
        test_runs = api.get_all_test_runs()
        self.assertEqual(1, len(test_runs))
        self.assertEqual(test_run_b.id, test_runs[0].id)
        self.assertEqual(1, len(api.get_test_run_metadata(test_run_b.id)))
        self.assertEqual(0, len(api.get_test_run_metadata(test_run_a.id)))

    def test_get_run_metadata(self):
        run_a = api.create_run()
        run_b = api.create_run()
        api.add_run_metadata({'not_a_key': 'not_a_value'}, run_b.id)
        a_metadata = api.get_run_metadata(run_a.uuid)
        b_metadata = api.get_run_metadata(run_b.uuid)
        self.assertEqual([], a_metadata)
        self.assertEqual(1, len(b_metadata))
        b_metadata = b_metadata[0].to_dict()
        self.assertEqual(run_b.id, b_metadata['run_id'])
        self.assertEqual('not_a_key', b_metadata['key'])
        self.assertEqual('not_a_value', b_metadata['value'])

    def test_get_runs_by_key_value(self):
        api.create_run()
        run_b = api.create_run()
        api.add_run_metadata({'not_a_key': 'not_a_value'}, run_b.id)
        found_runs = api.get_runs_by_key_value('not_a_key', 'not_a_value')
        self.assertEqual(1, len(found_runs))
        self.assertEqual(run_b.id, found_runs[0].id)
        self.assertEqual(run_b.uuid, found_runs[0].uuid)

    def test_get_tests_from_run_id(self):
        run_a = api.create_run()
        run_b = api.create_run()
        test_a = api.create_test('fake_test')
        test_b = api.create_test('fake_test2')
        api.create_test_run(test_a.id, run_a.id, 'fail',
                            start_time=datetime.datetime(1914, 6, 28, 10, 45,
                                                         0))
        api.create_test_run(test_a.id, run_b.id, 'fail',
                            start_time=datetime.datetime.utcnow())
        api.create_test_run(test_b.id, run_a.id, 'success',
                            start_time=datetime.datetime(1914, 6, 28, 10, 45,
                                                         0))
        result = api.get_tests_from_run_id(run_a.id)
        self.assertEqual(2, len(result))
        self.assertIn(test_a.id, [x.id for x in result])
        self.assertIn(test_a.test_id, [x.test_id for x in result])
        self.assertIn(test_b.id, [x.id for x in result])
        self.assertIn(test_b.test_id, [x.test_id for x in result])

    def test_get_all_runs_time_series_by_key(self):
        time_a = datetime.datetime(1914, 6, 28, 10, 45, 0)
        run_a = api.create_run(run_at=time_a)
        run_b = api.create_run()
        time_c = datetime.datetime(1918, 11, 11, 11, 11, 11)
        run_c = api.create_run(run_at=time_c)
        api.add_run_metadata({'not_a_key': 'not_a_value'}, run_b.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'c_value'}, run_c.id)
        result = api.get_all_runs_time_series_by_key('a_key')
        self.assertEqual(2, len(result.keys()))
        self.assertIn(time_a.date(), [x.date() for x in result.keys()])
        self.assertIn(time_c.date(), [x.date() for x in result.keys()])

    def test_get_recent_runs_by_key_value_metadata(self):
        run_a = api.create_run()
        run_b = api.create_run()
        run_c = api.create_run()
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_c.id)
        api.add_run_metadata({'a_key': 'b_value'}, run_b.id)
        result = api.get_recent_runs_by_key_value_metadata('a_key', 'a_value')
        self.assertEqual(2, len(result))
        self.assertIn(run_a.id, [x.id for x in result])
        self.assertNotIn(run_b.id, [x.id for x in result])
        self.assertIn(run_c.id, [x.id for x in result])

    def test_get_recent_runs_by_key_value_metadata_with_start_date(self):
        run_a = api.create_run(run_at=datetime.datetime(
            1914, 6, 28, 10, 45, 0))
        run_b = api.create_run()
        run_c = api.create_run()
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_c.id)
        api.add_run_metadata({'a_key': 'b_value'}, run_b.id)
        result = api.get_recent_runs_by_key_value_metadata(
            'a_key', 'a_value', start_date=datetime.datetime(
                1918, 11, 11, 11, 11, 11))
        self.assertEqual(1, len(result))
        self.assertNotIn(run_a.id, [x.id for x in result])
        self.assertNotIn(run_b.id, [x.id for x in result])
        self.assertIn(run_c.id, [x.id for x in result])

    def test_get_recent_runs_by_key_value_metadata_one_run(self):
        timestamp = datetime.datetime(1914, 6, 28, 10, 45, 0)
        run_a = api.create_run(run_at=timestamp)
        run_b = api.create_run()
        run_c = api.create_run()
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_c.id)
        api.add_run_metadata({'a_key': 'b_value'}, run_b.id)
        result = api.get_recent_runs_by_key_value_metadata('a_key', 'a_value',
                                                           num_runs=1)
        self.assertEqual(1, len(result))
        self.assertNotIn(run_a.id, [x.id for x in result])
        self.assertNotIn(run_b.id, [x.id for x in result])
        self.assertIn(run_c.id, [x.id for x in result])

    def test_get_test_runs_by_status_for_run_ids_no_meta(self):
        run_b = api.create_run(artifacts='fake_url')
        run_a = api.create_run()
        run_c = api.create_run()
        test_a = api.create_test('fake_test')
        start_timestamp = datetime.datetime(1914, 6, 28, 10, 45, 0)
        stop_timestamp = datetime.datetime(1914, 6, 28, 10, 50, 0)
        api.create_test_run(test_a.id, run_a.id, 'success',
                            datetime.datetime.utcnow())
        api.create_test_run(test_a.id, run_b.id, 'fail',
                            start_timestamp, stop_timestamp)
        api.create_test_run(test_a.id, run_c.id, 'success',
                            datetime.datetime.utcnow())
        result = api.get_test_runs_by_status_for_run_ids(
            'fail', [run_a.uuid, run_b.uuid, run_c.uuid])
        self.assertEqual(1, len(result))
        self.assertEqual({
            'test_id': u'fake_test',
            'link': u'fake_url',
            'start_time': start_timestamp,
            'stop_time': stop_timestamp,
        }, result[0])

    def test_get_test_runs_by_status_for_run_ids_with_meta(self):
        run_b = api.create_run(artifacts='fake_url')
        run_a = api.create_run()
        run_c = api.create_run()
        test_a = api.create_test('fake_test')
        api.add_run_metadata({'a_key': 'b'}, run_b.id)
        api.add_run_metadata({'a_key': 'a'}, run_a.id)
        api.add_run_metadata({'a_key': 'c'}, run_c.id)
        start_timestamp = datetime.datetime(1914, 6, 28, 10, 45, 0)
        stop_timestamp = datetime.datetime(1914, 6, 28, 10, 50, 0)
        api.create_test_run(test_a.id, run_a.id, 'success',
                            datetime.datetime.utcnow())
        api.create_test_run(test_a.id, run_b.id, 'fail',
                            start_timestamp, stop_timestamp)
        api.create_test_run(test_a.id, run_c.id, 'success',
                            datetime.datetime.utcnow())
        result = api.get_test_runs_by_status_for_run_ids(
            'fail', [run_a.uuid, run_b.uuid, run_c.uuid], key='a_key')
        self.assertEqual(1, len(result))
        self.assertEqual({
            'test_id': u'fake_test',
            'link': u'fake_url',
            'start_time': start_timestamp,
            'stop_time': stop_timestamp,
            'a_key': 'b',
        }, result[0])

    def test_get_test_runs_by_status_for_run_ids_with_run_id(self):
        run_b = api.create_run(artifacts='fake_url')
        run_a = api.create_run()
        run_c = api.create_run()
        test_a = api.create_test('fake_test')
        api.add_run_metadata({'a_key': 'b'}, run_b.id)
        api.add_run_metadata({'a_key': 'a'}, run_a.id)
        api.add_run_metadata({'a_key': 'c'}, run_c.id)
        start_timestamp = datetime.datetime(1914, 6, 28, 10, 45, 0)
        stop_timestamp = datetime.datetime(1914, 6, 28, 10, 50, 0)
        api.create_test_run(test_a.id, run_a.id, 'success',
                            datetime.datetime.utcnow())
        api.create_test_run(test_a.id, run_b.id, 'fail',
                            start_timestamp, stop_timestamp)
        api.create_test_run(test_a.id, run_c.id, 'success',
                            datetime.datetime.utcnow())
        result = api.get_test_runs_by_status_for_run_ids(
            'fail', [run_a.uuid, run_b.uuid, run_c.uuid], key='a_key',
            include_run_id=True)
        self.assertEqual(1, len(result))
        self.assertEqual({
            'test_id': u'fake_test',
            'link': u'fake_url',
            'start_time': start_timestamp,
            'stop_time': stop_timestamp,
            'a_key': 'b',
            'uuid': run_b.uuid,
        }, result[0])

    def test_get_all_runs_time_series_by_key_with_overlap(self):
        time_a = datetime.datetime(1914, 6, 28, 10, 45, 0)
        run_a = api.create_run(run_at=time_a)
        run_b = api.create_run()
        time_c = datetime.datetime(1918, 11, 11, 11, 11, 11)
        run_c = api.create_run(run_at=time_c)
        run_d = api.create_run(run_at=time_a)
        api.add_run_metadata({'not_a_key': 'not_a_value'}, run_b.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'c_value'}, run_c.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_d.id)
        result = api.get_all_runs_time_series_by_key('a_key')
        self.assertEqual(2, len(result.keys()))
        self.assertIn(time_a.date(), [x.date() for x in result.keys()])
        self.assertIn(time_c.date(), [x.date() for x in result.keys()])
        self.assertIn(time_a.date(), [x.date() for x in result.keys()])
        self.assertEqual(len(result[time_a]['a_value']), 2)

    def test_get_run_failure_rate_by_key_value_metadata(self):
        run_a = api.create_run(fails=100, passes=0)
        run_b = api.create_run()
        run_c = api.create_run(passes=100, fails=0)
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_c.id)
        api.add_run_metadata({'a_key': 'b_value'}, run_b.id)
        fail_rate = api.get_run_failure_rate_by_key_value_metadata(
            'a_key', 'a_value')
        self.assertEqual(50, fail_rate)

    def test_get_test_prefixes(self):
        api.create_test('prefix.token.token')
        api.create_test('setUpClass (prefix.token.token)')
        api.create_test('other.token.token')
        api.create_test('justonetoken')

        prefixes = api.get_test_prefixes()
        self.assertEqual(len(prefixes), 3)
        self.assertIn('prefix', prefixes)
        self.assertIn('other', prefixes)
        self.assertIn('justonetoken', prefixes)

    def test_get_tests_by_prefix(self):
        api.create_test('prefix.token.token')
        api.create_test('setUpClass (prefix.token.token)')
        api.create_test('other.token.token')
        api.create_test('justonetoken')

        tests = api.get_tests_by_prefix('prefix')
        self.assertEqual(len(tests), 2)
        self.assertIn('prefix.token.token', [test.test_id for test in tests])
        self.assertIn('setUpClass (prefix.token.token)',
                      [test.test_id for test in tests])

        tests = api.get_tests_by_prefix('other')
        self.assertEqual(len(tests), 1)
        self.assertIn('other.token.token', [test.test_id for test in tests])

        tests = api.get_tests_by_prefix('prefix', limit=1, offset=1)
        self.assertEqual(len(tests), 1)
        self.assertIn('setUpClass (prefix.token.token)',
                      [test.test_id for test in tests])

        tests = api.get_tests_by_prefix('justonetoken')
        self.assertEqual(len(tests), 1)
        self.assertIn('justonetoken', [test.test_id for test in tests])

    def test_update_test(self):
        create_res = api.create_test('fake_test')
        values = {'run_count': 2, 'run_time': 1.2}
        update_res = api.update_test(values, create_res.id)
        res = api.get_test_by_test_id('fake_test')
        self.assertEqual(res.id, update_res.id)
        self.assertEqual(res.test_id, 'fake_test')
        self.assertEqual(res.run_time, 1.2)
        self.assertEqual(res.run_count, 2)

    def test_update_test_run(self):
        run = api.create_run()
        test = api.create_test('fake_test')
        test_run = api.create_test_run(test.id, run.id, 'fail')
        start_time = datetime.datetime.utcnow()
        stop_time = datetime.datetime.utcnow()
        start_time = start_time.replace(microsecond=0)
        stop_time = start_time.replace(microsecond=0)
        values = {'status': 'success', 'start_time': start_time,
                  'stop_time': stop_time}
        update_test_run = api.update_test_run(values, test_run.id)
        all_test_runs = api.get_all_test_runs()
        self.assertEqual(len(all_test_runs), 1)
        self.assertEqual(update_test_run.id, all_test_runs[0].id)
        self.assertEqual(all_test_runs[0].status, 'success')
        self.assertEqual(all_test_runs[0].start_time, start_time)
        self.assertEqual(all_test_runs[0].stop_time, stop_time)

    def test_get_test_run_by_id(self):
        run = api.create_run()
        test = api.create_test('fake_test')
        test_run = api.create_test_run(test.id, run.id, 'fail')
        res = api.get_test_run_by_id(test_run.id)
        self.assertEqual(res.id, test_run.id)
        self.assertEqual(res.status, 'fail')
        self.assertEqual(res.run_id, run.id)

    def test_get_test_by_id(self):
        create_res = api.create_test('fake_test', 2, 1, 1, 1.2)
        res = api.get_test_by_id(create_res.id)
        self.assertEqual(res.id, create_res.id)
        self.assertEqual(res.test_id, 'fake_test')
        self.assertEqual(res.run_time, 1.2)
        self.assertEqual(res.run_count, 2)

    def test_get_run_id_from_uuid(self):
        run = api.create_run()
        run_id = api.get_run_id_from_uuid(run.uuid)
        self.assertEqual(run_id, run.id)

    def test_get_run_by_id(self):
        run = api.create_run()
        res = api.get_run_by_id(run.id)
        self.assertEqual(res.id, run.id)
        self.assertEqual(res.uuid, run.uuid)

    def test_get_runs_by_ids(self):
        run_a = api.create_run()
        run_b = api.create_run()
        res = api.get_runs_by_ids([run_a.id, run_b.id])
        self.assertIn(run_a.id, [x.id for x in res])
        self.assertIn(run_b.id, [x.id for x in res])
        self.assertIn(run_a.uuid, [x.uuid for x in res])
        self.assertIn(run_b.uuid, [x.uuid for x in res])

    def test_get_test_runs_by_test_id(self):
        run_a = api.create_run()
        run_b = api.create_run()
        run_c = api.create_run()
        test = api.create_test('fake_test')
        api.create_test_run(test.id, run_a.id, 'success')
        api.create_test_run(test.id, run_b.id, 'fail')
        api.create_test_run(test.id, run_c.id, 'success')
        res = api.get_test_runs_by_test_id(test.id)
        self.assertEqual(3, len(res))
        self.assertIn(run_a.id, [x.run_id for x in res])
        self.assertIn(run_b.id, [x.run_id for x in res])
        self.assertIn(run_c.id, [x.run_id for x in res])

    def test_get_test_metadata(self):
        test = api.create_test('fake_test')
        test_meta = {
            'test_a': 'a',
            'test_b': 'b',
            'test_c': 'c',
        }
        api.add_test_metadata(test_meta, test.id)
        test_metadata = api.get_test_metadata(test.id)
        self.assertEqual(3, len(test_metadata))
        for meta in test_metadata:
            self.assertIn(meta.key, test_meta.keys())
            self.assertIn(meta.value, test_meta.values())

    def test_get_recent_failed_runs_by_run_metadata_no_start_date(self):
        run_a = api.create_run(fails=1)
        api.create_run()
        run_c = api.create_run(fails=2)
        api.add_run_metadata({'fake_key': 'fake_value'}, run_a.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_c.id)
        results = api.get_recent_failed_runs_by_run_metadata('zeon', 'zaku')
        self.assertEqual(1, len(results))
        self.assertEqual(run_c.id, results[0].id)

    def test_get_recent_failed_runs_by_run_metadata_with_start_date(self):
        run_a = api.create_run(fails=1)
        api.create_run()
        run_c = api.create_run(fails=2)
        run_d = api.create_run(fails=3,
                               run_at=datetime.datetime(1914, 6, 28,
                                                        10, 45, 0))
        api.add_run_metadata({'fake_key': 'fake_value'}, run_a.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_c.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_d.id)
        results = api.get_recent_failed_runs_by_run_metadata(
            'zeon', 'zaku', start_date=datetime.date(1970, 1, 1))
        self.assertEqual(1, len(results))
        self.assertEqual(run_c.id, results[0].id)

    def test_get_runs_count(self):
        for i in range(0, 100):
            api.create_run()
        count = api.get_runs_count()
        self.assertEqual(100, count)

    def test_get_runs_count_zero(self):
        self.assertEqual(0, api.get_runs_count())

    def test_get_runs_counts_by_run_metadata_no_start_date(self):
        run_a = api.create_run(fails=1)
        api.create_run()
        run_c = api.create_run(fails=2)
        api.add_run_metadata({'fake_key': 'fake_value'}, run_a.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_c.id)
        results = api.get_runs_counts_by_run_metadata('zeon', 'zaku')
        self.assertEqual(1, results)
        results = api.get_runs_counts_by_run_metadata('zeon', 'dom')
        self.assertEqual(0, results)

    def test_get_runs_counts_by_run_metadata_with_start_date(self):
        run_a = api.create_run(fails=1)
        api.create_run()
        run_c = api.create_run(fails=2)
        run_d = api.create_run(fails=3,
                               run_at=datetime.datetime(1914, 6, 28,
                                                        10, 45, 0))
        api.add_run_metadata({'fake_key': 'fake_value'}, run_a.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_c.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_d.id)
        results = api.get_runs_counts_by_run_metadata(
            'zeon', 'zaku', start_date=datetime.date(1970, 1, 1))
        self.assertEqual(1, results)
        results = api.get_runs_counts_by_run_metadata(
            'zeon', 'zaku', start_date=datetime.date(1914, 6, 27))
        self.assertEqual(2, results)
        results = api.get_runs_counts_by_run_metadata(
            'zeon', 'dom', start_date=datetime.date(1914, 6, 27))
        self.assertEqual(0, results)

    def test_get_test_run_metadata(self):
        run = api.create_run()
        test = api.create_test('fake_test')
        test_run = api.create_test_run(test.id, run.id, 'success',
                                       datetime.datetime.utcnow(),
                                       datetime.datetime.utcnow())
        run_meta = {
            'key_a': 'value_b',
            'key_b': 'value_a',
        }
        api.add_test_run_metadata(run_meta, test_run.id)
        test_run_metadata = api.get_test_run_metadata(test_run.id)
        self.assertEqual(2, len(test_run_metadata))
        for meta in test_run_metadata:
            self.assertIn(meta.key, run_meta.keys())
            self.assertIn(meta.value, run_meta.values())

    def test_get_all_runs_by_date(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=20)
        api.create_run(run_at=timestamp_a)
        api.create_run(run_at=timestamp_b)
        api.create_run(run_at=timestamp_c)
        res = api.get_all_runs_by_date()
        self.assertEqual(3, len(res))
        self.assertIn(timestamp_a, [x.run_at for x in res])
        self.assertIn(timestamp_b, [x.run_at for x in res])
        self.assertIn(timestamp_c, [x.run_at for x in res])
        res = api.get_all_runs_by_date(start_date=timestamp_c)
        self.assertEqual(1, len(res))
        self.assertEqual(timestamp_c, res[0].run_at)
        res = api.get_all_runs_by_date(stop_date=timestamp_a)
        self.assertEqual(1, len(res))
        self.assertEqual(timestamp_a, res[0].run_at)
        res = api.get_all_runs_by_date(start_date=timestamp_a,
                                       stop_date=timestamp_b)
        self.assertEqual(2, len(res))
        self.assertIn(timestamp_a, [x.run_at for x in res])
        self.assertIn(timestamp_b, [x.run_at for x in res])

    def test_get_latest_run(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=20)
        api.create_run(run_at=timestamp_a)
        api.create_run(run_at=timestamp_b)
        api.create_run(run_at=timestamp_c)
        res = api.get_latest_run()
        self.assertEqual(timestamp_c, res.run_at)

    def test_get_failing_from_run(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=20)
        run_a = api.create_run()
        run_b = api.create_run()
        test_a = api.create_test('fake_test')
        test_b = api.create_test('fake_test2')
        api.create_test_run(test_a.id, run_a.id, 'fail',
                            start_time=timestamp_a)
        api.create_test_run(test_a.id, run_b.id, 'fail',
                            start_time=timestamp_a)
        api.create_test_run(test_b.id, run_b.id, 'fail',
                            start_time=timestamp_b)
        api.create_test_run(test_b.id, run_a.id, 'success',
                            start_time=timestamp_c)
        res = api.get_failing_from_run(run_a.id)
        self.assertEqual(1, len(res))
        self.assertEqual(test_a.id, res[0].test_id)
        self.assertEqual(timestamp_a, res[0].start_time)
        res = api.get_failing_from_run(run_b.id)
        self.assertEqual(2, len(res))
        self.assertEqual('fail', res[0].status)
        self.assertIn(timestamp_a, [x.start_time for x in res])
        self.assertIn(timestamp_b, [x.start_time for x in res])

    def test_get_test_run_time_series(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=15)
        run_a = api.create_run()
        run_b = api.create_run()
        test = api.create_test('fake_test')
        api.create_test_run(test.id, run_a.id, 'success',
                            start_time=timestamp_a,
                            end_time=timestamp_b)
        api.create_test_run(test.id, run_b.id, 'success',
                            start_time=timestamp_b,
                            end_time=timestamp_c)
        res = api.get_test_run_time_series(test.id)
        self.assertEqual(res[timestamp_a], 600)
        self.assertEqual(res[timestamp_b], 300)

    def test_get_test_status_time_series(self):
        timestamp_a = datetime.datetime.utcnow().replace(microsecond=0)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=20)
        run_a = api.create_run()
        run_b = api.create_run()
        test = api.create_test('fake_test')
        api.create_test_run(test.id, run_a.id, 'success',
                            start_time=timestamp_a,
                            end_time=timestamp_b)
        api.create_test_run(test.id, run_b.id, 'fail',
                            start_time=timestamp_b,
                            end_time=timestamp_c)
        res = api.get_test_status_time_series(test.id)
        self.assertEqual(res[timestamp_a], 'success')
        self.assertEqual(res[timestamp_b], 'fail')

    def test_get_recent_failed_runs(self):
        run_a = api.create_run(fails=1)
        api.create_run()
        run_c = api.create_run(fails=2)
        run_d = api.create_run(fails=1)
        res = api.get_recent_failed_runs()
        self.assertEqual(3, len(res))
        self.assertIn(run_a.uuid, res)
        self.assertIn(run_c.uuid, res)
        self.assertIn(run_d.uuid, res)

    def test_get_recent_failed_runs_with_start_date(self):
        api.create_run(fails=1, run_at=datetime.datetime(
            1914, 6, 28, 10, 45, 0))
        api.create_run()
        run_c = api.create_run(fails=2)
        run_d = api.create_run(fails=1)
        res = api.get_recent_failed_runs(start_date=datetime.datetime(
            1918, 11, 11, 11, 11, 11))
        self.assertEqual(2, len(res))
        self.assertIn(run_c.uuid, res)
        self.assertIn(run_d.uuid, res)

    def test_get_recent_successful_runs(self):
        run_a = api.create_run(passes=1)
        run_b = api.create_run()
        run_c = api.create_run(passes=2)
        api.create_run(fails=1)
        res = api.get_recent_successful_runs()
        self.assertEqual(3, len(res))
        self.assertIn(run_a.uuid, res)
        self.assertIn(run_b.uuid, res)
        self.assertIn(run_c.uuid, res)

    def test_get_recent_successful_runs_with_start_date(self):
        api.create_run(passes=1, run_at=datetime.datetime(
            1914, 6, 28, 10, 45, 0))
        run_b = api.create_run()
        run_c = api.create_run(passes=2)
        api.create_run(fails=1)
        res = api.get_recent_successful_runs(start_date=datetime.datetime(
            1918, 11, 11, 11, 11, 11))
        self.assertEqual(2, len(res))
        self.assertIn(run_b.uuid, res)
        self.assertIn(run_c.uuid, res)

    def test_get_test_counts_in_date_range_as_str(self):
        timestamp_str_a = 'Dec 01 2015'
        timestamp_str_b = 'Dec 20 2015'
        timestamp_a = datetime.datetime(2015, 12, 2, 10, 00, 00)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=20)
        timestamp_d = datetime.datetime(2015, 12, 22, 10, 00, 00)
        run_a = api.create_run()
        run_b = api.create_run()
        run_c = api.create_run()
        run_d = api.create_run()
        test = api.create_test('fake_test')
        api.create_test_run(test.id, run_a.id, 'success',
                            timestamp_a, timestamp_b)
        api.create_test_run(test.id, run_b.id, 'fail',
                            timestamp_a, timestamp_c)
        api.create_test_run(test.id, run_c.id, 'success',
                            timestamp_a, timestamp_d)
        api.create_test_run(test.id, run_d.id, 'skip',
                            timestamp_c, timestamp_d)
        res = api.get_test_counts_in_date_range(test.id,
                                                timestamp_str_a,
                                                timestamp_str_b)
        self.assertEqual(1, res['success'])
        self.assertEqual(1, res['failure'])
        self.assertEqual(0, res['skips'])

    def test_get_test_counts_in_date_range(self):
        timestamp_a = datetime.datetime(2015, 12, 2, 10, 00, 00)
        timestamp_b = timestamp_a + datetime.timedelta(minutes=10)
        timestamp_c = timestamp_a + datetime.timedelta(minutes=20)
        timestamp_d = datetime.datetime(2015, 12, 22, 10, 00, 00)
        timerange_a = datetime.datetime(2015, 12, 1)
        timerange_b = datetime.datetime(2015, 12, 20)
        run_a = api.create_run()
        run_b = api.create_run()
        run_c = api.create_run()
        run_d = api.create_run()
        test = api.create_test('fake_test')
        api.create_test_run(test.id, run_a.id, 'success',
                            timestamp_a, timestamp_b)
        api.create_test_run(test.id, run_b.id, 'fail',
                            timestamp_a, timestamp_c)
        api.create_test_run(test.id, run_c.id, 'success',
                            timestamp_a, timestamp_d)
        api.create_test_run(test.id, run_d.id, 'skip',
                            timestamp_c, timestamp_d)
        res = api.get_test_counts_in_date_range(test.id,
                                                timerange_a,
                                                timerange_b)
        self.assertEqual(1, res['success'])
        self.assertEqual(1, res['failure'])
        self.assertEqual(0, res['skips'])

    def test_get_failing_test_ids_from_runs_by_key_value(self):
        test_a = api.create_test('fake_test')
        test_b = api.create_test('fake_test1')
        test_c = api.create_test('fake_test2')
        test_d = api.create_test('fake_test3')
        run_a = api.create_run()
        run_b = api.create_run()
        run_c = api.create_run()
        api.add_run_metadata({'a_key': 'a_value'}, run_a.id)
        api.add_run_metadata({'a_key': 'a_value'}, run_b.id)
        api.add_run_metadata({'a_key': 'b_value'}, run_c.id)
        api.create_test_run(test_a.id, run_a.id, 'fail')
        api.create_test_run(test_b.id, run_b.id, 'success')
        api.create_test_run(test_c.id, run_b.id, 'fail')
        api.create_test_run(test_d.id, run_c.id, 'fail')
        res = api.get_failing_test_ids_from_runs_by_key_value('a_key',
                                                              'a_value')
        self.assertEqual(2, len(res))
        self.assertIn(test_a.test_id, res)
        self.assertIn(test_c.test_id, res)

    def test_add_test_run_attachments(self):
        test = api.create_test('fake_test')
        run = api.create_run()
        test_run = api.create_test_run(test.id, run.id, 'success')
        attach_dict = {'attach_label': b'attach',
                       'attach_label_a': b'attach_a'}
        res = api.add_test_run_attachments(attach_dict, test_run.id)
        self.assertEqual(2, len(res))
        self.assertEqual(test_run.id, res[0].test_run_id)
        self.assertIn('attach_label', [x.label for x in res])
        self.assertIn('attach_label_a', [x.label for x in res])
        self.assertIn(b'attach', [x.attachment for x in res])
        self.assertIn(b'attach_a', [x.attachment for x in res])

    def test_get_ids_for_all_tests(self):
        test_a = api.create_test('fake_test')
        test_b = api.create_test('fake_test1')
        test_c = api.create_test('fake_test2')
        res = api.get_ids_for_all_tests()
        self.assertIsInstance(res, types.GeneratorType)
        res = list(res)
        self.assertIn((test_c.id,), res)
        self.assertIn((test_b.id,), res)
        self.assertIn((test_a.id,), res)

    def test_get_recent_successful_runs_by_run_metadata_no_start_date(self):
        run_a = api.create_run(fails=0, passes=1)
        api.create_run()
        run_c = api.create_run(fails=0, passes=2)
        api.add_run_metadata({'fake_key': 'fake_value'}, run_a.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_c.id)
        results = api.get_recent_successful_runs_by_run_metadata('zeon',
                                                                 'zaku')
        self.assertEqual(1, len(results))
        self.assertEqual(run_c.id, results[0].id)

    def test_get_recent_successful_runs_by_run_metadata_with_start_date(self):
        run_a = api.create_run(fails=0, passes=1)
        api.create_run()
        run_c = api.create_run(fails=0, passes=2)
        run_d = api.create_run(fails=0, passes=3,
                               run_at=datetime.datetime(1914, 6, 28,
                                                        10, 45, 0))
        api.add_run_metadata({'fake_key': 'fake_value'}, run_a.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_c.id)
        api.add_run_metadata({'zeon': 'zaku'}, run_d.id)
        results = api.get_recent_successful_runs_by_run_metadata(
            'zeon', 'zaku', start_date=datetime.date(1970, 1, 1))
        self.assertEqual(1, len(results))
        self.assertEqual(run_c.id, results[0].id)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path
import tempfile

from oslo_config import cfg
from oslo_config import fixture

from subunit2sql.targets import localdir
from subunit2sql.tests import base


class TestLocaldir(base.TestCase):
    def setUp(self):
        super(TestLocaldir, self).setUp()
        self.useFixture(fixture.Config(cfg.CONF))
        self.tdir = tempfile.mkdtemp()
        cfg.CONF.set_override(name='attachments_storage_dir',
                              override=self.tdir)
        self.ad = localdir.AttachmentsDir()

    def test_localdir_enabled_when_configured(self):
        self.assertTrue(self.ad.enabled())

    def test_localdir_disabled_when_no_conf(self):
        cfg.CONF.clear_override(name='attachments_storage_dir')
        self.assertFalse(self.ad.enabled())

    def test_localdir_status_ignores_non_attachments(self):
        self.ad.status(test_id='foo.test',
                       test_status='melancholy')
        self.ad.stopTestRun()
        self.assertEqual(0, len(os.listdir(self.tdir)))

    def test_localdir_saves_testless_attachments(self):
        self.ad.status(file_name='super.txt',
                       file_bytes=b'the quick brown fox',
                       route_code='routecode1')
        self.ad.status(file_name='super.txt',
                       file_bytes=b'jumped over the lazy brown dog',
                       route_code='routecode2')
        self.ad.stopTestRun()
        expected_path = os.path.join(self.tdir, 'routecode1', 'super.txt')
        self.assertTrue(os.path.exists(expected_path))
        with open(expected_path, 'rb') as f:
            self.assertEqual(b'the quick brown fox', f.read())
        expected_path = os.path.join(self.tdir, 'routecode2', 'super.txt')
        self.assertTrue(os.path.exists(expected_path))
        with open(expected_path, 'rb') as f:
            self.assertEqual(b'jumped over the lazy brown dog', f.read())



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import shutil
import subprocess

import fixtures as fix
from oslo_concurrency.fixture import lockutils as lock_fixture
from oslo_concurrency import lockutils
from oslo_config import fixture as config_fixture
from oslo_db import options
from six.moves.urllib import parse as urlparse

from subunit2sql.db import api as session
from subunit2sql.migrations import cli
from subunit2sql.tests import db_test_utils

DB_SCHEMA = ""


def execute_cmd(cmd=None):
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT, shell=True)
    output = proc.communicate()[0]
    if proc.returncode != 0:
        raise Exception('Command failed with output:\n%s' % output)
    return output


class Database(fix.Fixture):
    def _cache_schema(self):
        global DB_SCHEMA
        if not DB_SCHEMA:
            db_test_utils.run_migration("head")

    def cleanup(self):
        engine = session.get_engine()
        engine.dispose()

    def reset(self):
        self._cache_schema()
        engine = session.get_engine()
        engine.dispose()
        engine.connect()

    def setUp(self):
        super(Database, self).setUp()
        self.reset()
        self.addCleanup(self.cleanup)


class MySQLConfFixture(config_fixture.Config):
    """Fixture to manage global conf settings."""
    def _drop_db(self):
        addr = urlparse.urlparse(self.url)
        database = addr.path.strip('/')
        loc_pieces = addr.netloc.split('@')
        host = loc_pieces[1]
        auth_pieces = loc_pieces[0].split(':')
        user = auth_pieces[0]
        password = ""
        if len(auth_pieces) > 1:
            if auth_pieces[1].strip():
                password = "-p\"%s\"" % auth_pieces[1]
        sql = ("drop database if exists %(database)s; create "
               "database %(database)s;") % {'database': database}
        cmd = ("mysql -u \"%(user)s\" %(password)s -h %(host)s "
               "-e \"%(sql)s\"") % {'user': user, 'password': password,
                                    'host': host, 'sql': sql}
        execute_cmd(cmd)

    def setUp(self):
        super(MySQLConfFixture, self).setUp()
        self.register_opts(options.database_opts, group='database')
        self.url = db_test_utils.get_connect_string("mysql")
        self.set_default('connection', self.url, group='database')
        lockutils.set_defaults(lock_path='/tmp')
        self._drop_db()


class PostgresConfFixture(config_fixture.Config):
    """Fixture to manage global conf settings."""
    def _drop_db(self):
        addr = urlparse.urlparse(self.url)
        database = addr.path.strip('/')
        loc_pieces = addr.netloc.split('@')
        host = loc_pieces[1]

        auth_pieces = loc_pieces[0].split(':')
        user = auth_pieces[0]
        password = ""
        if len(auth_pieces) > 1:
            password = auth_pieces[1].strip()
        pg_file = os.path.join(os.path.expanduser('~'), '.pgpass')
        if os.path.isfile(pg_file):
            tmp_path = os.path.join('/tmp', 'pgpass')
            shutil.move(pg_file, tmp_path)
            self.addCleanup(shutil.move, tmp_path, pg_file)

        pg_pass = '*:*:*:%(user)s:%(password)s' % {
            'user': user, 'password': password}
        with open(pg_file, 'w') as fd:
            fd.write(pg_pass)
        os.chmod(pg_file, 384)
        # note(boris-42): We must create and drop database, we can't
        # drop database which we have connected to, so for such
        # operations there is a special database template1.
        sqlcmd = ('psql -w -U %(user)s -h %(host)s -c'
                  ' "%(sql)s" -d template1')

        # NOTE(masayukig): We terminate sessions because some closed
        # sessions are remaining until here
        sql = ("select pg_terminate_backend(pg_stat_activity.pid) "
               "from pg_stat_activity "
               "where pg_stat_activity.datname = '%(database)s';")
        sql = sql % {'database': database}
        term_session = sqlcmd % {'user': user, 'host': host,
                                 'sql': sql}
        execute_cmd(term_session)

        sql = ("drop database if exists %(database)s;")
        sql = sql % {'database': database}
        droptable = sqlcmd % {'user': user, 'host': host,
                              'sql': sql}
        execute_cmd(droptable)
        sql = ("create database %(database)s;")
        sql = sql % {'database': database}
        createtable = sqlcmd % {'user': user, 'host': host,
                                'sql': sql}
        execute_cmd(createtable)

    def setUp(self):
        super(PostgresConfFixture, self).setUp()
        self.register_opts(options.database_opts, group='database')
        self.register_opts(cli.MIGRATION_OPTS)
        self.url = db_test_utils.get_connect_string("postgres")
        self.set_default('connection', self.url, group='database')
        self.set_default('disable_microsecond_data_migration', False)
        lockutils.set_defaults(lock_path='/tmp')
        self._drop_db()


class SqliteConfFixture(config_fixture.Config):
    """Fixture to manage global conf settings."""
    def _drop_db(self):
        if os.path.exists(db_test_utils.SQLITE_TEST_DATABASE_PATH):
            os.remove(db_test_utils.SQLITE_TEST_DATABASE_PATH)

    def setUp(self):
        super(SqliteConfFixture, self).setUp()
        self.register_opts(options.database_opts, group='database')
        self.register_opts(cli.MIGRATION_OPTS)
        self.url = db_test_utils.get_connect_string("sqlite")
        self.set_default('connection', self.url, group='database')
        self.set_default('disable_microsecond_data_migration', False)
        lockutils.set_defaults(lock_path='/tmp')
        self._drop_db()
        self.addCleanup(self.cleanup)

    def cleanup(self):
        self._drop_db()


class LockFixture(lock_fixture.LockFixture):
    def __init__(self, name):
        lockutils.set_defaults(lock_path='/tmp')
        super(LockFixture, self).__init__(name, 'subunit-db-lock-')



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

import matplotlib
import matplotlib.dates as dates
import matplotlib.pyplot as plt
from oslo_config import cfg
import pandas as pd

from subunit2sql.db import api

CONF = cfg.CONF

matplotlib.style.use('ggplot')


def set_cli_opts(parser):
    parser.add_argument('--dcmd_key', default='build_queue',
                        help="When running dailycount, the runs included are "
                             "filtered based on metadata. If one is not "
                             "provided, a default of \"build_queue\" is used "
                             "as the metadata key.")
    parser.add_argument('--dcmd_value', default='gate',
                        help="When running dailycount, the runs included are "
                             "filtered based on metadata. This option "
                             "specifies the value that metadata should have. "
                             "If not specified, a default of \"gate\" is "
                             "used.")


def generate_series():
    if CONF.start_date:
        start_date = datetime.datetime.strptime(CONF.start_date, '%Y-%m-%d')
    else:
        start_date = None
    if CONF.stop_date:
        stop_date = datetime.datetime.strptime(CONF.stop_date, '%Y-%m-%d')
    else:
        stop_date = None
    session = api.get_session()
    test_starts = api.get_test_run_series(start_date=start_date,
                                          stop_date=stop_date,
                                          session=session,
                                          key=CONF.command.dcmd_key,
                                          value=CONF.command.dcmd_value)
    session.close()
    ts = pd.Series(test_starts)
    daily_count = ts.resample('D').sum().fillna(value=0)
    mean = daily_count.rolling(window=10, center=False).mean()
    rolling_std = daily_count.rolling(window=10, center=False).std()
    plt.figure()
    title = CONF.title or 'Number of Tests run Daily'
    plt.title(title)
    plt.ylabel('Number of tests')
    fig, ax = plt.subplots(1)
    fig.autofmt_xdate()
    plt.title(title)
    plt.ylabel('Number of tests')
    xfmt = dates.DateFormatter("%b %d %Y")
    ax.xaxis_date()
    ax.xaxis.set_major_formatter(xfmt)

    plt.plot(daily_count.index[10:], daily_count[10:], 'k',
             label='Daily Test Count')
    plt.plot(mean.index[10:], mean[10:], 'b', label='Avg. Daily Test Count')
    upper_std_dev = mean + 2 * rolling_std
    lower_std_dev = mean - 2 * rolling_std
    # Set negative numbers to 0
    lower_std_dev[lower_std_dev < 0] = 0
    plt.fill_between(rolling_std.index[10:], lower_std_dev[10:],
                     upper_std_dev[10:],
                     color='b', alpha=0.2, label='Std Dev')
    plt.legend()
    plt.savefig(CONF.output, dpi=CONF.dpi)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import mock
from oslo_db.sqlalchemy import test_migrations
import sqlalchemy
import testscenarios

from subunit2sql.db import models
from subunit2sql.tests import base
from subunit2sql.tests import db_test_utils
from subunit2sql.tests import subunit2sql_fixtures as fixtures

load_tests = testscenarios.load_tests_apply_scenarios


class TestModelsMigrations(test_migrations.ModelsMigrationsSync,
                           base.TestCase):
    """Test for checking of equality models state and migrations.

    Output is a list that contains information about differences between db and
    models. Output example::
       [('add_table',
         Table('bat', MetaData(bind=None),
               Column('info', String(), table=<bat>), schema=None)),
        ('remove_table',
         Table(u'bar', MetaData(bind=None),
               Column(u'data', VARCHAR(), table=<bar>), schema=None)),
        ('add_column',
         None,
         'foo',
         Column('data', Integer(), table=<foo>)),
        ('remove_column',
         None,
         'foo',
         Column(u'old_data', VARCHAR(), table=None)),
        [('modify_nullable',
          None,
          'foo',
          u'x',
          {'existing_server_default': None,
          'existing_type': INTEGER()},
          True,
          False)]]
    * ``remove_*`` means that there is extra table/column/constraint in db;
    * ``add_*`` means that it is missing in db;
    * ``modify_*`` means that on column in db is set wrong
      type/nullable/server_default. Element contains information:
        - what should be modified,
        - schema,
        - table,
        - column,
        - existing correct column parameters,
        - right value,
        - wrong value.
    """

    scenarios = [
        ('mysql', {'dialect': 'mysql'}),
        ('postgresql', {'dialect': 'postgres'}),
        ('sqlite', {'dialect': 'sqlite'})
    ]
    # NOTE(mtreinish): Mock out db attr because oslo.db tries to assert
    # control of how the opportunistic db is setup. But, since subunit2sql's
    # tests already do this, just let oslo.db think it's doing the operations
    db = mock.MagicMock()

    def setUp(self):
        super(TestModelsMigrations, self).setUp()
        self.useFixture(fixtures.LockFixture(self.dialect))
        if not db_test_utils.is_backend_avail(self.dialect):
            raise self.skipTest('%s is not available' % self.dialect)
        if self.dialect == 'sqlite':
            raise self.skipException('sqlite skipped because of model sync '
                                     'issue with BigInteger vs Integer')
        if self.dialect == 'mysql':
            self.useFixture(fixtures.MySQLConfFixture())
        elif self.dialect == 'postgres':
            self.useFixture(fixtures.PostgresConfFixture())
        connect_string = db_test_utils.get_connect_string(self.dialect)
        self.engine = sqlalchemy.create_engine(connect_string)

    def get_engine(self):
        return self.engine

    def get_metadata(self):
        return models.BASE.metadata

    def db_sync(self, engine):
        db_test_utils.run_migration('head', engine)

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table' and name == 'alembic_version':
                return False
        return super(TestModelsMigrations, self).include_object(
            object_, name, type_, reflected, compare_to)

    def filter_metadata_diff(self, diff):
        return list(filter(self.remove_unrelated_errors, diff))

    def remove_unrelated_errors(self, element):
        insp = sqlalchemy.engine.reflection.Inspector.from_engine(
            self.get_engine())
        dialect = self.get_engine().dialect.name
        if isinstance(element, tuple):
            if dialect == 'mysql' and element[0] == 'remove_index':
                table_name = element[1].table.name
                for fk in insp.get_foreign_keys(table_name):
                    if fk['name'] == element[1].name:
                        return False
                cols = [c.name for c in element[1].expressions]
                for col in cols:
                    if col in insp.get_pk_constraint(
                            table_name)['constrained_columns']:
                        return False
        else:
            for modified, _, table, column, _, _, new in element:
                if modified == 'modify_default' and dialect == 'mysql':
                    constrained = insp.get_pk_constraint(table)
                    if column in constrained['constrained_columns']:
                        return False
        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import os.path
import tempfile

from oslo_config import cfg
import testtools

OPTIONS = [
    cfg.StrOpt('attachments_storage_dir', default=None,
               help='Any file attachments will be stored here')
]

cfg.CONF.register_cli_opts(OPTIONS)
cfg.CONF.register_opts(OPTIONS)


class AttachmentsDir(testtools.StreamResult):
    @classmethod
    def enabled(self):
        if not cfg.CONF.attachments_storage_dir:
            return False
        if not os.path.isdir(cfg.CONF.attachments_storage_dir):
            return False
        return True

    def __init__(self):
        self.pending_files = {}

    def status(self, test_id=None, test_status=None, test_tags=None,
               runnable=True, file_name=None, file_bytes=None, eof=False,
               mime_type=None, route_code=None, timestamp=None):
        if not file_name or not file_bytes:
            return
        target_dirs = []
        if route_code:
            target_dirs.append(route_code)
        if test_id:
            target_dirs.append(test_id)
        target_dir = os.path.join(cfg.CONF.attachments_storage_dir,
                                  *target_dirs)
        os.makedirs(target_dir)
        target_file_path = os.path.join(target_dir, file_name)
        if target_file_path not in self.pending_files:
            self.pending_files[target_file_path] = tempfile.NamedTemporaryFile(
                prefix='.{}'.format(file_name), dir=target_dir, delete=False)
        self.pending_files[target_file_path].write(file_bytes)

    def stopTestRun(self):
        for target_path, f in self.pending_files.items():
            f.close()
            os.rename(f.name, target_path)



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

import mock

from subunit2sql.tests import base
from subunit2sql import write_subunit

timestamp_a = datetime.datetime.utcnow()
timestamp_b = datetime.datetime.utcnow()


class TestWriteSubunit(base.TestCase):

    @mock.patch('subunit2sql.db.api.get_tests_run_dicts_from_run_id',
                return_value={'fake_test_id': {
                              'start_time': timestamp_a,
                              'stop_time': timestamp_b,
                              'status': 'success'}})
    @mock.patch('subunit2sql.write_subunit.write_test')
    @mock.patch('subunit2sql.db.api.get_session')
    def test_test_runs_without_metdata(self, api_mock, write_mock,
                                       sesion_mock):
        out = mock.MagicMock()
        write_subunit.sql2subunit('fake_id', out)
        self.assertEqual(write_mock.call_count, 1)
        args = write_mock.call_args_list[0][0]
        self.assertEqual(timestamp_a, args[1])
        self.assertEqual(timestamp_b, args[2])
        self.assertEqual('success', args[3])
        self.assertEqual('fake_test_id', args[4])
        self.assertEqual(None, args[5])

    @mock.patch('subunit2sql.db.api.get_tests_run_dicts_from_run_id',
                return_value={'fake_test_id': {
                    'start_time': timestamp_a,
                    'stop_time': timestamp_b,
                    'status': 'success',
                    'metadata': {
                        'key': 'value'}}})
    @mock.patch('subunit2sql.write_subunit.write_test')
    @mock.patch('subunit2sql.db.api.get_session')
    def test_test_runs_with_metdata(self, api_mock, write_mock, session_mock):
        out = mock.MagicMock()
        write_subunit.sql2subunit('fake_id', out)
        self.assertEqual(write_mock.call_count, 1)
        args = write_mock.call_args_list[0][0]
        self.assertEqual(timestamp_a, args[1])
        self.assertEqual(timestamp_b, args[2])
        self.assertEqual('success', args[3])
        self.assertEqual('fake_test_id', args[4])
        self.assertEqual({'key': 'value'}, args[5])



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add Attachments Table

Revision ID: 487f279b8c78
Revises: 1679b5bc102c
Create Date: 2015-05-27 15:18:21.653867

"""

# revision identifiers, used by Alembic.
revision = '487f279b8c78'
down_revision = '1679b5bc102c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('attachments',
                    sa.Column('id', sa.String(36), primary_key=True,
                              nullable=False),
                    sa.Column('test_run_id', sa.String(36),
                              sa.ForeignKey('test_runs.id'), nullable=False),
                    sa.Column('label', sa.String(255)),
                    sa.Column('attachment', sa.LargeBinary()),
                    mysql_engine='InnoDB')


def downgrade():
    op.drop_table('attachments')



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import operator

from dateutil import parser as date_parser
import matplotlib.pyplot as plt
import numpy
from oslo_config import cfg

from subunit2sql.db import api

CONF = cfg.CONF


def set_cli_opts(parser):
    parser.add_argument('key', nargs='?',
                        help='The run metadata key to aggregate the run times'
                             ' on.')
    parser.add_argument('--num', default=10, type=int,
                        help='The number of results to show. If 0 is set all '
                             'results will be shown')


def generate_series():
    session = api.get_session()
    start_date = None
    stop_date = None
    if CONF.start_date:
        start_date = date_parser.parse(CONF.start_date)
    if CONF.stop_date:
        stop_date = date_parser.parse(CONF.stop_date)
    ci_time = {}
    ci_time_temp = {}
    project_run_times = api.get_run_times_grouped_by_run_metadata_key(
        CONF.command.key, start_date=start_date, stop_date=stop_date,
        session=session)
    for project in project_run_times:
        ci_time_temp[project] = numpy.sum(project_run_times[project])
    sorted_times = sorted(ci_time_temp.items(), key=operator.itemgetter(1),
                          reverse=True)
    if CONF.command.num:
        sorted_times = sorted_times[:CONF.command.num]
    for project, time in sorted_times:
        ci_time[project] = time

    title = CONF.title or 'Aggregate Run Time grouped by %s' % CONF.command.key
    session.close()
    plt.bar(range(len(ci_time)), ci_time.values(), align='center', width=.1)
    plt.xticks(range(len(ci_time)), ci_time.keys(), rotation=90, fontsize=8)
    plt.title(title)
    plt.tight_layout()
    plt.savefig(CONF.output, dpi=CONF.dpi)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import uuid

from oslo_db.sqlalchemy import models  # noqa
import six
import sqlalchemy as sa
from sqlalchemy.ext import declarative

BASE = declarative.declarative_base()


class SubunitBase(models.ModelBase):
    """Base class for Subunit Models."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __table_initialized__ = False

    def save(self, session=None):
        from subunit2sql.db import api as db_api
        super(SubunitBase, self).save(session or db_api.get_session())

    def keys(self):
        return list(self.__dict__.keys())

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def to_dict(self):
        d = self.__dict__.copy()
        d.pop("_sa_instance_state")
        return d


class Test(BASE, SubunitBase):
    __tablename__ = 'tests'
    __table_args__ = (sa.Index('ix_test_ids', 'id', 'test_id',
                               mysql_length={'test_id': 30}),
                      sa.Index('ix_tests_test_id', 'test_id',
                               mysql_length=30))
    id = sa.Column(sa.BigInteger, primary_key=True)
    test_id = sa.Column(sa.String(256),
                        nullable=False)
    run_count = sa.Column(sa.Integer())
    success = sa.Column(sa.Integer())
    failure = sa.Column(sa.Integer())
    run_time = sa.Column(sa.Float())


class Run(BASE, SubunitBase):
    __tablename__ = 'runs'
    __table_args__ = (sa.Index('ix_run_at', 'run_at'),
                      sa.Index('ix_run_uuid', 'uuid'))
    uuid = sa.Column(sa.String(36),
                     default=lambda: six.text_type(uuid.uuid4()))
    id = sa.Column(sa.BigInteger, primary_key=True)
    skips = sa.Column(sa.Integer())
    fails = sa.Column(sa.Integer())
    passes = sa.Column(sa.Integer())
    run_time = sa.Column(sa.Float())
    artifacts = sa.Column(sa.Text())
    run_at = sa.Column(sa.DateTime,
                       default=datetime.datetime.utcnow)


class TestRun(BASE, SubunitBase):
    __tablename__ = 'test_runs'
    __table_args__ = (sa.Index('ix_test_id_status', 'test_id', 'status'),
                      sa.Index('ix_test_id_start_time', 'test_id',
                               'start_time'),
                      sa.Index('ix_test_runs_test_id', 'test_id'),
                      sa.Index('ix_test_runs_run_id', 'run_id'),
                      sa.Index('ix_test_runs_start_time', 'start_time'),
                      sa.Index('ix_test_runs_stop_time', 'stop_time'),
                      sa.UniqueConstraint('test_id', 'run_id',
                                          name='uq_test_runs'))

    id = sa.Column(sa.BigInteger, primary_key=True)
    test_id = sa.Column(sa.BigInteger)
    run_id = sa.Column(sa.BigInteger)
    status = sa.Column(sa.String(256))
    start_time = sa.Column(sa.DateTime())
    start_time_microsecond = sa.Column(sa.Integer(), default=0)
    stop_time = sa.Column(sa.DateTime())
    stop_time_microsecond = sa.Column(sa.Integer(), default=0)
    test = sa.orm.relationship(Test, backref=sa.orm.backref('test_run_test'),
                               foreign_keys=test_id,
                               primaryjoin=test_id == Test.id)
    run = sa.orm.relationship(Run, backref=sa.orm.backref('test_run_run'),
                              foreign_keys=run_id,
                              primaryjoin=run_id == Run.id)


class RunMetadata(BASE, SubunitBase):
    __tablename__ = 'run_metadata'
    __table_args__ = (sa.Index('ix_run_key_value', 'key', 'value'),
                      sa.Index('ix_run_id', 'run_id'),
                      sa.UniqueConstraint('run_id', 'key', 'value',
                                          name='uq_run_metadata'))

    id = sa.Column(sa.BigInteger, primary_key=True)
    key = sa.Column(sa.String(255))
    value = sa.Column(sa.String(255))
    run_id = sa.Column(sa.BigInteger)
    run = sa.orm.relationship(Run, backref='run', foreign_keys=run_id,
                              primaryjoin=run_id == Run.id)


class TestRunMetadata(BASE, SubunitBase):
    __tablename__ = 'test_run_metadata'
    __table_args__ = (sa.Index('ix_test_run_key_value', 'key', 'value'),
                      sa.Index('ix_test_run_id', 'test_run_id'),
                      sa.UniqueConstraint('test_run_id', 'key', 'value',
                                          name='uq_test_run_metadata'))

    id = sa.Column(sa.BigInteger, primary_key=True)
    key = sa.Column(sa.String(255))
    value = sa.Column(sa.String(255))
    test_run_id = sa.Column(sa.BigInteger)
    test_run = sa.orm.relationship(TestRun,
                                   backref=sa.orm.backref('test_run_meta'),
                                   foreign_keys=test_run_id,
                                   primaryjoin=test_run_id == TestRun.id)


class TestMetadata(BASE, SubunitBase):
    __tablename__ = 'test_metadata'
    __table_args__ = (sa.Index('ix_test_key_value', 'key', 'value'),
                      sa.Index('ix_test_id', 'test_id'),
                      sa.UniqueConstraint('test_id', 'key', 'value',
                                          name='uq_test_metadata'))

    id = sa.Column(sa.BigInteger, primary_key=True)
    key = sa.Column(sa.String(255))
    value = sa.Column(sa.String(255))
    test_id = sa.Column(sa.BigInteger)
    test = sa.orm.relationship(Test, backref='test', foreign_keys=test_id,
                               primaryjoin=test_id == Test.id)


class Attachments(BASE, SubunitBase):
    __tablename__ = 'attachments'
    __table_args__ = (sa.Index('ix_attach_test_run_id', 'test_run_id'),)
    id = sa.Column(sa.BigInteger, primary_key=True)
    test_run_id = sa.Column(sa.BigInteger)
    label = sa.Column(sa.String(255))
    attachment = sa.Column(sa.LargeBinary())
    test_run = sa.orm.relationship(TestRun, backref='test_run_attach',
                                   foreign_keys=test_run_id,
                                   primaryjoin=test_run_id == TestRun.id)



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Populate run_time for existing tests

Revision ID: 5332fe255095
Revises: 28ac1ba9c3db
Create Date: 2014-10-03 10:23:25.469128

"""

# revision identifiers, used by Alembic.
revision = '5332fe255095'
down_revision = '28ac1ba9c3db'


from oslo_db.sqlalchemy import utils as db_utils

from subunit2sql.db import api as db_api
from subunit2sql.db import models
from subunit2sql import read_subunit


def upgrade():
    query = db_utils.model_query(
        models.Test, db_api.get_session()).filter(
            models.Test.success > 0, models.Test.run_time == None).join(
                models.TestRun,
                models.Test.id == models.TestRun.test_id).filter_by(
                    status='success').values(models.Test.id,
                                             models.TestRun.start_time,
                                             models.TestRun.stop_time)

    results = {}
    for test_run in query:
        delta = read_subunit.get_duration(test_run[1], test_run[2])
        if test_run[0] in results:
            results[test_run[0]].append(delta)
        else:
            results[test_run[0]] = [delta]

    for test in results:
        avg = float(sum(results[test])) / float(len(results[test]))
        db_api.update_test({'run_time': avg}, test)


def downgrade():
    # NOTE(mtreinish) there is no possible downgrade for this migration, since
    # we won't be able to tell which rows had run_time NULL before this.
    # Ideally this would have been baked into 163fd5aa1380 and the downgrade
    # there of deleting the column would have covered this. But, because that
    # wasn't included as a part of the released migration we can't change it.
    pass



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import datetime
import sys

from oslo_config import cfg
import subunit
from subunit import iso8601

from subunit2sql.db import api
from subunit2sql import shell

STATUS_CODES = frozenset([
    'exists',
    'fail',
    'skip',
    'success',
    'uxsuccess',
    'xfail',
])

CONF = cfg.CONF

SHELL_OPTS = [
    cfg.StrOpt('run_id', positional=True,
               help='Run id to use for creating a subunit stream'),
    cfg.StrOpt('out_path', short='o', default=None,
               help='Path to write the subunit stream output, if none '
                    'is specified STDOUT will be used'),
    cfg.BoolOpt('average', short='a', default=False,
                help='Generate a subunit stream for all the rows in the tests '
                     'table using the average run_time for the duration.')
]


def cli_opts():
    for opt in SHELL_OPTS:
        cfg.CONF.register_cli_opt(opt)


def convert_datetime(timestamp):
    tz_timestamp = timestamp.replace(tzinfo=iso8601.UTC)
    return tz_timestamp


def write_test(output, start_time, stop_time, status, test_id, metadatas):
    write_status = output.status
    kwargs = {}
    if metadatas:
        if 'tags' in metadatas:
            tags = metadatas['tags']
            kwargs['test_tags'] = tags.split(',')
        if 'attrs' in metadatas:
            test_id = test_id + '[' + metadatas['attrs'] + ']'
    start_time = convert_datetime(start_time)
    kwargs['timestamp'] = start_time
    kwargs['test_id'] = test_id
    write_status(**kwargs)
    if status in STATUS_CODES:
        kwargs['test_status'] = status
        kwargs['timestamp'] = convert_datetime(stop_time)
    write_status(**kwargs)


def sql2subunit(run_id, output=sys.stdout):
    session = api.get_session()
    test_runs = api.get_tests_run_dicts_from_run_id(run_id, session)
    session.close()
    output = subunit.v2.StreamResultToBytes(output)
    output.startTestRun()
    for test_id in test_runs:
        test = test_runs[test_id]
        # NOTE(mtreinish): test_run_metadata is not guaranteed to be present
        # for the test_run.
        metadata = test.get('metadata', None)
        write_test(output, test['start_time'], test['stop_time'],
                   test['status'], test_id, metadata)
    output.stopTestRun()


def avg_sql2subunit(output=sys.stdout):
    session = api.get_session()
    tests = api.get_all_tests(session=session)
    session.close()
    output = subunit.v2.StreamResultToBytes(output)
    output.startTestRun()
    for test in tests:
        if not test.run_time:
            continue
        start_time = datetime.datetime.now()
        stop_time = start_time + datetime.timedelta(0, test.run_time)
        write_test(output, start_time, stop_time, 'success', test.test_id, [])
    output.stopTestRun()


def list_opts():
    opt_list = copy.deepcopy(SHELL_OPTS)
    return [('DEFAULT', opt_list)]


def main():
    cli_opts()
    shell.parse_args(sys.argv)
    if not CONF.run_id and not CONF.average:
        print('You must specify either a run_id or generate an average run'
              ' stream')
        return 1
    if CONF.run_id and CONF.average:
        print('You can either generate a stream for a run_id or an average run'
              ' stream, but not both.')
        return 1
    if CONF.out_path:
        fd = open(CONF.out_path, 'w')
    else:
        fd = sys.stdout
    if not CONF.average:
        sql2subunit(CONF.run_id, fd)
    else:
        avg_sql2subunit(fd)
    if CONF.out_path:
        fd.close()

if __name__ == "__main__":
    sys.exit(main())



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class Subunit2SQLException(Exception):
    """Base Subunit2SQL Exception.

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    message = "An unknown exception occurred"

    def __init__(self, *args, **kwargs):
        super(Subunit2SQLException, self).__init__()
        try:
            self._error_string = self.message % kwargs
        except Exception:
            # at least get the core message out if something happened
            self._error_string = self.message
        if len(args) > 0:
            # If there is a non-kwarg parameter, assume it's the error
            # message or reason description and tack it on to the end
            # of the exception message
            # Convert all arguments into their string representations...
            args = ["%s" % arg for arg in args]
            self._error_string = (self._error_string +
                                  "\nDetails: %s" % '\n'.join(args))

    def __str__(self):
        return self._error_string


class InvalidRunCount(Subunit2SQLException):
    message = "Invalid Run Count"


class UnknownStatus(Subunit2SQLException):
    message = "Unknown test status"



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
import pandas as pd

from subunit2sql.analysis import utils
from subunit2sql.db import api

CONF = cfg.CONF


def set_cli_opts(parser):
    parser.add_argument('test_id', nargs='?',
                        help='Test id to extract time series for. This '
                             'previously took a UUID from the tests.id '
                             'column, however this will no longer work. It '
                             'only works with a value from tests.test_id.'),
    parser.add_argument('--success-graph', action='store_true',
                        help='Also graph successes'),
    parser.add_argument('--skip-graph', action='store_true',
                        help='Also graph skips')


def generate_series():
    session = api.get_session()
    test_id = api.get_id_from_test_id(CONF.command.test_id, session)
    test_statuses = api.get_test_status_time_series(test_id, session)
    if not CONF.title:
        test = api.get_test_by_id(test_id, session)
    session.close()
    ts = pd.Series(test_statuses)
    ts = utils.filter_dates(ts)
    run_count = len(ts)
    if run_count == 0:
        print("Query returned no data.")
        exit(-1)
    failures = ts[ts.isin(['fail', 'unxsuccess'])]
    successes = ts[ts.isin(['success', 'xfail'])]
    skips = ts[ts.isin(['skip'])]
    fail_count = len(failures)
    success_count = len(successes)
    skip_count = len(skips)
    fail_group = failures.groupby(failures.index.date).agg(len)
    success_group = successes.groupby(successes.index.date).agg(len)
    skip_group = skips.groupby(skips.index.date).agg(len)
    if not CONF.title:
        plot = fail_group.plot().set_title(test.test_id)
    else:
        plot = fail_group.plot().set_title(CONF.title)
    if CONF.command.success_graph:
        if success_count:
            success_group.plot()
    if CONF.command.skip_graph:
        if skip_count:
            skip_group.plot()

    def percent(count, total):
        count = float(count)
        total = float(total)
        return (count / total) * 100.0

    print('Fail Percentage: %.4f%%' % percent(fail_count, run_count))
    print('Success Percentage: %.4f%%' % percent(success_count, run_count))
    print('Skip Percentage: %.4f%%' % percent(skip_count, run_count))
    fig = plot.get_figure()
    fig.savefig(CONF.output, dpi=CONF.dpi)
    return ts



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Create avg runtime column in test table

Revision ID: 163fd5aa1380
Revises: 3db7b49816d5
Create Date: 2014-06-16 15:45:19.221576

"""

# revision identifiers, used by Alembic.
revision = '163fd5aa1380'
down_revision = '3db7b49816d5'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('tests', sa.Column('run_time', sa.Float(), nullable=True))


def downgrade():
    op.drop_column('tests', 'run_time')



# Copyright 2015 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""uuid to integer ids

Revision ID: 2822a408bdd0
Revises: b96122f780
Create Date: 2015-10-14 14:18:18.820521

"""

# revision identifiers, used by Alembic.
revision = '2822a408bdd0'
down_revision = 'b96122f780'

from alembic import context
from alembic import op
import sqlalchemy as sa


def upgrade():
    migration_context = context.get_context()
    if migration_context.dialect.name == 'sqlite':
        new_id_type = sa.Integer
    else:
        new_id_type = sa.BigInteger
    # Create new tests table
    op.create_table('tests_new',
                    sa.Column('id', sa.String(36)),
                    sa.Column('new_id', new_id_type, primary_key=True,
                              autoincrement=True, nullable=False),
                    sa.Column('test_id', sa.String(256), nullable=False),
                    sa.Column('run_count', sa.Integer()),
                    sa.Column('success', sa.Integer()),
                    sa.Column('failure', sa.Integer()),
                    sa.Column('run_time', sa.Float(), nullable=True),
                    mysql_engine='InnoDB')

    # Create new test_runs table
    op.create_table('test_runs_new',
                    sa.Column('id', sa.String(36)),
                    sa.Column('new_id', new_id_type, primary_key=True,
                              nullable=False),
                    sa.Column('test_id', sa.String(36), nullable=False),
                    sa.Column('new_test_id', new_id_type),
                    sa.Column('run_id', sa.String(36), nullable=False),
                    sa.Column('new_run_id', new_id_type),
                    sa.Column('status', sa.String(256)),
                    sa.Column('start_time', sa.DateTime()),
                    sa.Column('stop_time', sa.DateTime()),
                    sa.Column('start_time_microsecond', sa.Integer(),
                              default=0),
                    sa.Column('stop_time_microsecond', sa.Integer(),
                              default=0),
                    mysql_engine='InnoDB')

    # Create new runs table
    op.create_table('runs_new',
                    sa.Column('id', sa.String(36)),
                    sa.Column('new_id', new_id_type(), primary_key=True,
                              nullable=False, autoincrement=True),
                    sa.Column('skips', sa.Integer()),
                    sa.Column('fails', sa.Integer()),
                    sa.Column('passes', sa.Integer()),
                    sa.Column('run_time', sa.Float()),
                    sa.Column('artifacts', sa.Text()),
                    sa.Column('run_at', sa.DateTime()),
                    mysql_engine='InnoDB')

    # Create new run_metadata table
    op.create_table('run_metadata_new',
                    sa.Column('id', sa.String(36),
                              nullable=False),
                    sa.Column('new_id', new_id_type(), nullable=False,
                              primary_key=True, autoincrement=True),
                    sa.Column('key', sa.String(255)),
                    sa.Column('value', sa.String(255)),
                    sa.Column('run_id', sa.String(36), nullable=False),
                    sa.Column('new_run_id', new_id_type),
                    mysql_engine='InnoDB')

    # Create new test runs metadata table
    op.create_table('test_run_metadata_new',
                    sa.Column('id', sa.String(36), nullable=False),
                    sa.Column('new_id', new_id_type, primary_key=True,
                              nullable=False, autoincrement=True),
                    sa.Column('key', sa.String(255)),
                    sa.Column('value', sa.String(255)),
                    sa.Column('test_run_id', sa.String(36),
                              nullable=False),
                    sa.Column('new_test_run_id', new_id_type),
                    mysql_engine='InnoDB')

    # Create new test metadata table
    op.create_table('test_metadata_new',
                    sa.Column('id', sa.String(36),
                              nullable=False),
                    sa.Column('new_id', new_id_type, primary_key=True,
                              nullable=False, autoincrement=True),
                    sa.Column('key', sa.String(255)),
                    sa.Column('value', sa.String(255)),
                    sa.Column('test_id', sa.String(36), nullable=False),
                    sa.Column('new_test_id', new_id_type),
                    mysql_engine='InnoDB')
    # Create new tests attachments table
    op.create_table('attachments_new',
                    sa.Column('id', sa.String(36), nullable=False),
                    sa.Column('new_id', new_id_type, primary_key=True,
                              nullable=False, autoincrement=True),
                    sa.Column('test_run_id', sa.String(36), nullable=False),
                    sa.Column('new_test_run_id', new_id_type),
                    sa.Column('label', sa.String(255)),
                    sa.Column('attachment', sa.LargeBinary()),
                    mysql_engine='InnoDB')

    # Now populate new tables
    if migration_context.dialect.name == 'postgresql':
        key_word = 'key'
    else:
        key_word = '`key`'
    op.execute('INSERT INTO tests_new (id, test_id, run_count, success, '
               'failure, run_time) SELECT id, test_id, run_count, success, '
               'failure, run_time FROM tests')
    op.execute('INSERT INTO runs_new (id, skips, fails, passes, run_time, '
               'artifacts, run_at) SELECT id, skips, fails, passes, run_time, '
               'artifacts, run_at FROM runs')
    op.execute('INSERT INTO test_runs_new (id, test_id, new_test_id, run_id, '
               'new_run_id, status, start_time, stop_time, '
               'start_time_microsecond, stop_time_microsecond) SELECT tr.id, '
               'tr.test_id, tn.new_id, tr.run_id, rn.new_id, status, '
               'start_time, stop_time, start_time_microsecond, '
               'stop_time_microsecond FROM test_runs tr INNER JOIN runs_new '
               'rn ON rn.id = tr.run_id INNER JOIN tests_new tn '
               'ON tn.id=tr.test_id')
    op.execute('INSERT INTO test_metadata_new (id, {}, value, test_id, '
               'new_test_id) SELECT tm.id, tm.key, tm.value, tm.test_id, '
               'tn.new_id FROM test_metadata tm INNER JOIN tests_new tn '
               'ON tn.id = tm.test_id'.format(key_word))
    op.execute('INSERT INTO test_run_metadata_new (id, {}, value, '
               'test_run_id, new_test_run_id) SELECT trm.id, trm.key, '
               'trm.value, trm.test_run_id, trn.new_id FROM test_run_metadata '
               'trm INNER JOIN test_runs_new trn ON trm.test_run_id = '
               'trn.id'.format(key_word))
    op.execute('INSERT INTO attachments_new (id, test_run_id, '
               'new_test_run_id, label, attachment) SELECT a.id, '
               'a.test_run_id, trn.new_id, a.label, a.attachment FROM '
               'attachments a INNER JOIN test_runs_new trn '
               'ON a.test_run_id = trn.id')
    op.execute('INSERT INTO run_metadata_new (id, {}, value, run_id, '
               'new_run_id) SELECT rm.id, rm.key, rm.value, rm.run_id, '
               'rn.new_id FROM run_metadata rm INNER JOIN runs_new rn '
               'ON rm.run_id = rn.id'.format(key_word))

    # Switch columns
    if migration_context.dialect.name == 'postgresql':

        op.drop_column('attachments_new', 'id')
        op.alter_column('attachments_new', 'new_id', new_column_name='id',
                        existing_type=new_id_type,
                        autoincrement=True)
        op.drop_column('attachments_new', 'test_run_id')
        op.alter_column('attachments_new', 'new_test_run_id',
                        new_column_name='test_run_id')
        op.drop_column('test_run_metadata_new', 'id')
        op.alter_column('test_run_metadata_new', 'new_id',
                        new_column_name='id',
                        existing_type=new_id_type,
                        autoincrement=True)
        op.drop_column('test_run_metadata_new', 'test_run_id')
        op.alter_column('test_run_metadata_new', 'new_test_run_id',
                        new_column_name='test_run_id',
                        existing_type=new_id_type)
        op.drop_column('run_metadata_new', 'id')
        op.alter_column('run_metadata_new', 'new_id',
                        new_column_name='id')
        op.drop_column('run_metadata_new', 'run_id')
        op.alter_column('run_metadata_new', 'new_run_id',
                        new_column_name='run_id',
                        existing_type=new_id_type)
        op.drop_column('test_metadata_new', 'id')
        op.alter_column('test_metadata_new', 'new_id',
                        new_column_name='id')
        op.drop_column('test_metadata_new', 'test_id')
        op.alter_column('test_metadata_new', 'new_test_id',
                        new_column_name='test_id',
                        existing_type=new_id_type)
        op.drop_column('test_runs_new', 'id')
        op.alter_column('test_runs_new', 'new_id',
                        new_column_name='id')
        op.drop_column('test_runs_new', 'test_id')
        op.alter_column('test_runs_new', 'new_test_id',
                        new_column_name='test_id',
                        existing_type=new_id_type)
        op.drop_column('test_runs_new', 'run_id')
        op.alter_column('test_runs_new', 'new_run_id',
                        new_column_name='run_id',
                        existing_type=new_id_type)
        op.drop_column('tests_new', 'id')
        op.alter_column('tests_new', 'new_id',
                        new_column_name='id')
        op.alter_column('runs_new', 'id',
                        new_column_name='uuid')
        op.alter_column('runs_new', 'new_id',
                        new_column_name='id')
    else:
        # http://dev.mysql.com/doc/refman/5.7/en/innodb-online-ddl-syntax.html
        # Ths is a specific workaround for limited tmpdir space in the
        # OpenStack infra MySQL server. With old_alter_table=OFF, mysql creates
        # temporary files that are very large while building the new table.
        # So generally, while the old method is less desirable for concurrency,
        # it is safer, and we don't need online DDL since this migration
        # uses a _new table anyway.
        if migration_context.dialect.name == 'mysql':
            op.execute('SET SESSION old_alter_table=ON')
        with op.batch_alter_table("attachments_new") as batch_op:
            batch_op.drop_column('id')
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)
            batch_op.drop_column('test_run_id')
            batch_op.alter_column('new_test_run_id',
                                  new_column_name='test_run_id',
                                  existing_type=new_id_type)
        with op.batch_alter_table("test_run_metadata_new") as batch_op:
            batch_op.drop_column('id')
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)
            batch_op.drop_column('test_run_id')
            batch_op.alter_column('new_test_run_id',
                                  new_column_name='test_run_id',
                                  existing_type=new_id_type)
        with op.batch_alter_table("run_metadata_new") as batch_op:
            batch_op.drop_column('id')
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)
            batch_op.drop_column('run_id')
            batch_op.alter_column('new_run_id', new_column_name='run_id',
                                  existing_type=new_id_type)
        with op.batch_alter_table("test_metadata_new") as batch_op:
            batch_op.drop_column('id')
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)
            batch_op.drop_column('test_id')
            batch_op.alter_column('new_test_id', new_column_name='test_id',
                                  existing_type=new_id_type)
        with op.batch_alter_table("test_runs_new") as batch_op:
            batch_op.drop_column('id')
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)
            batch_op.drop_column('test_id')
            batch_op.alter_column('new_test_id', new_column_name='test_id',
                                  existing_type=new_id_type)
            batch_op.drop_column('run_id')
            batch_op.alter_column('new_run_id', new_column_name='run_id',
                                  existing_type=new_id_type)
        with op.batch_alter_table("tests_new") as batch_op:
            batch_op.drop_column('id')
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)
        with op.batch_alter_table("runs_new") as batch_op:
            batch_op.alter_column('id', new_column_name='uuid',
                                  existing_type=sa.VARCHAR(36))
            batch_op.alter_column('new_id', new_column_name='id',
                                  primary_key=True,
                                  nullable=False,
                                  existing_type=new_id_type,
                                  autoincrement=True)

    # Sanity checks
    errors = []
    for table in ('tests', 'runs', 'test_runs', 'test_metadata',
                  'test_run_metadata', 'run_metadata', 'attachments'):
        old_count = op.get_bind().execute(
            "SELECT COUNT(id) FROM {}".format(table)).first()[0]
        new_count = op.get_bind().execute(
            "SELECT COUNT(id) FROM {}_new".format(table)).first()[0]
        if old_count != new_count:
            errors.append("{} has {} rows and {}_new has {} rows".format(
                table, old_count, table, new_count))
    if errors:
        raise RuntimeError("Failed count checks: {}".format(','.join(errors)))

    # Rename tables
    op.rename_table("tests", "tests_old")
    op.rename_table("runs", "runs_old")
    op.rename_table("test_runs", "test_runs_old")
    op.rename_table("test_metadata", "test_metadata_old")
    op.rename_table("test_run_metadata", "test_run_metadata_old")
    op.rename_table("run_metadata", "run_metadata_old")
    op.rename_table("attachments", "attachments_old")
    op.rename_table("tests_new", "tests")
    op.rename_table("runs_new", "runs")
    op.rename_table("test_runs_new", "test_runs")
    op.rename_table("test_metadata_new", "test_metadata")
    op.rename_table("test_run_metadata_new", "test_run_metadata")
    op.rename_table("run_metadata_new", "run_metadata")
    op.rename_table("attachments_new", "attachments")

    # Drop olds
    op.drop_table("test_run_metadata_old")
    op.drop_table("attachments_old")
    op.drop_table("test_metadata_old")
    op.drop_table("run_metadata_old")
    op.drop_table("test_runs_old")
    op.drop_table("runs_old")
    op.drop_table("tests_old")

    # Create indexes -- sqlite keeps the old ones around for some reason
    if migration_context.dialect.name != 'sqlite':
        op.create_index('ix_test_ids', 'tests', ['id', 'test_id'],
                        mysql_length={'test_id': 30})
        op.create_index('ix_test_key_value', 'test_metadata',
                        ['key', 'value'])
        op.create_index('ix_test_run_key_value', 'test_run_metadata',
                        ['key', 'value'])
        op.create_index('ix_run_key_value', 'run_metadata',
                        ['key', 'value'])
        op.create_index('ix_test_id_status', 'test_runs',
                        ['test_id', 'status'], mysql_length={'status': 30})
        op.create_index('ix_test_id_start_time', 'test_runs',
                        ['test_id', 'start_time'])
        op.create_unique_constraint('uq_test_runs', 'test_runs',
                                    ['test_id', 'run_id'])
    op.create_index('ix_run_uuid', 'runs', ['uuid'])
    op.create_index('ix_tests_test_id', 'tests', ['test_id'], mysql_length=30)
    op.create_index('ix_test_runs_test_id', 'test_runs', ['test_id'])
    op.create_index('ix_test_runs_run_id', 'test_runs', ['run_id'])
    op.create_index('ix_test_runs_start_time', 'test_runs', ['start_time'])
    op.create_index('ix_test_runs_stop_time', 'test_runs', ['stop_time'])


def downgrade():
    raise NotImplementedError()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dateutil import parser as date_parser
from oslo_config import cfg
import pandas as pd

from subunit2sql.db import api

CONF = cfg.CONF


def set_cli_opts(parser):
    parser.add_argument('test_ids', nargs='*',
                        help='Test ids to graph counts for, if none are '
                             'specified all tests will be used')
    parser.add_argument('--no-success-graph', action='store_true',
                        help='Do not graph successes'),
    parser.add_argument('--skip-graph', action='store_true',
                        help='Also graph skips')


def generate_series():
    session = api.get_session()
    test_dict = {}
    if not CONF.start_date and not CONF.stop_date:
        tests = api.get_all_tests(session)
        for test in tests:
            if CONF.command.test_ids:
                if test.test_id in CONF.command.test_ids:
                    test_dict[test.test_id] = {
                        'success': int(test.success),
                        'failure': int(test.failure),
                    }
            else:
                test_dict[test.test_id] = {
                    'success': int(test.success),
                    'failure': int(test.failure),
                }
    else:
        start_date = None
        stop_date = None
        if CONF.start_date:
            start_date = date_parser.parse(CONF.start_date)
        if CONF.stop_date:
            stop_date = date_parser.parse(CONF.stop_date)
        if CONF.command.test_ids:
            ids = [api.get_id_from_test_id(x) for x in CONF.command.test_ids]
        else:
            ids = api.get_ids_for_all_tests(session)
        for test in ids:
            test_dict[test] = api.get_test_counts_in_date_range(
                test, start_date, stop_date, session)
    if CONF.command.no_success_graph:
        for test in test_dict:
            test_dict[test].pop('success')
    if CONF.command.skip_graph:
        for test in test_dict:
            if not test_dict[test].get('skips'):
                test_id = api.get_id_from_test_id(test)
                test_dict[test]['skips'] = api.get_skip_counts(test_id)
    session.close()
    if not CONF.title:
        title = "Test status counts"
    else:
        title = CONF.title
    df = pd.DataFrame.from_dict(test_dict, orient='index')
    plot = df.plot(kind='barh', stacked=True).set_title(title)
    fig = plot.get_figure()
    fig.savefig(CONF.output, dpi=CONF.dpi)



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import fixtures
import testtools


class TestCase(testtools.TestCase):

    true = ('True', 'true', '1', 'yes')

    def setUp(self):
        super(TestCase, self).setUp()
        if os.environ.get('OS_STDOUT_CAPTURE') in self.true:
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if os.environ.get('OS_STDERR_CAPTURE') in self.true:
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))
        if (os.environ.get('OS_LOG_CAPTURE') != 'False' and
            os.environ.get('OS_LOG_CAPTURE') != '0'):
            self.useFixture(fixtures.LoggerFixture(nuke_handlers=False,
                                                   level=None))



# Copyright (c) 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Add missing fk indexes

Revision ID: 35cd45895e56
Revises: 2822a408bdd0
Create Date: 2015-11-30 15:32:37.218171

"""

# revision identifiers, used by Alembic.
revision = '35cd45895e56'
down_revision = '2822a408bdd0'

from alembic import context
from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    migration_context = context.get_context()
    insp = reflection.Inspector(migration_context.bind)
    test_run_meta_indx = insp.get_indexes('test_run_metadata')
    run_meta_indx = insp.get_indexes('run_metadata')
    test_meta_indx = insp.get_indexes('test_metadata')
    runs_indx = insp.get_indexes('runs')
    attach_indx = insp.get_indexes('attachments')
    if 'run_id' not in [
        x['column_names'][0] for x in run_meta_indx if len(x) == 1]:
        op.create_index('ix_run_id', 'run_metadata', ['run_id'])
    if 'test_id' not in [
        x['column_names'][0] for x in test_meta_indx if len(x) == 1]:
        op.create_index('ix_test_id', 'test_metadata', ['test_id'])
    if 'test_run_id' not in [
        x['column_names'][0] for x in test_run_meta_indx if len(x) == 1]:
        op.create_index('ix_test_run_id', 'test_run_metadata', ['test_run_id'])
    if 'run_at' not in [
        x['column_names'][0] for x in runs_indx if len(x) == 1]:
        op.create_index('ix_run_at', 'runs', ['run_at'])
    if 'test_run_id' not in [
        x['column_names'][0] for x in attach_indx if len(x) == 1]:
        op.create_index('ix_attach_test_run_id', 'attachments',
                        ['test_run_id'])


def downgrade():
    NotImplementedError()



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement
from alembic import context
from logging.config import fileConfig  # noqa

from subunit2sql.db import api as db_api


# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
subunit2sql_config = config.subunit2sql_config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    kwargs = dict()
    if subunit2sql_config.database.connection:
        kwargs['url'] = subunit2sql_config.database.connection
    elif subunit2sql_config.database.engine:
        kwargs['dialect_name'] = subunit2sql_config.database.engine
    else:
        kwargs['url'] = config.get_main_option("sqlalchemy.url")
    kwargs['target_metadata'] = target_metadata
    kwargs['render_as_batch'] = True
    context.configure(**kwargs)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """

    facade = db_api._create_facade_lazily()
    engine = facade.get_engine()
    connection = engine.connect()
    facade.get_sessionmaker().configure(bind=connection)

    context.configure(connection=connection,
                      target_metadata=target_metadata,
                      render_as_batch=True)

    try:
        with context.begin_transaction():
            context.run_migrations()
    finally:
        connection.close()
        facade.get_sessionmaker().configure(bind=engine)

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add indexes on common search fields

Revision ID: 1ff737bef438
Revises: 487f279b8c78
Create Date: 2015-06-25 14:27:10.465845

"""

# revision identifiers, used by Alembic.
revision = '1ff737bef438'
down_revision = '487f279b8c78'

from alembic import context
from alembic import op
from sqlalchemy.engine import reflection


def upgrade():
    migration_context = context.get_context()
    insp = reflection.Inspector(migration_context.bind)
    run_indx = insp.get_indexes('runs')
    run_indx_names = [x['name'] for x in run_indx]
    test_run_indx = insp.get_indexes('test_runs')
    test_run_indx_names = [x['name'] for x in test_run_indx]
    test_run_metad_indx = insp.get_indexes('test_run_metadata')
    test_run_metad_indx_names = [x['name'] for x in test_run_metad_indx]
    run_metad_indx = insp.get_indexes('run_metadata')
    run_metad_indx_names = [x['name'] for x in run_metad_indx]
    test_metad_indx = insp.get_indexes('test_metadata')
    test_metad_indx_names = [x['name'] for x in test_metad_indx]

    # Add indexes to time columns these are often used for searches and filters
    if 'ix_test_start_time' not in test_run_indx_names:
        op.create_index('ix_test_start_time', 'test_runs',
                        ['start_time'])
    if 'ix_test_stop_time' not in test_run_indx_names:
        op.create_index('ix_test_stop_time', 'test_runs',
                        ['stop_time'])
    if 'ix_run_at' not in run_indx_names:
        op.create_index('ix_run_at', 'runs', ['run_at'])
    # Add compound index on metadata tables key, value columns
    if 'ix_run_key_value' not in run_metad_indx_names:
        op.create_index('ix_run_key_value', 'run_metadata', ['key', 'value'])
    if 'ix_test_run_key_value' not in test_run_metad_indx_names:
        op.create_index('ix_test_run_key_value', 'test_run_metadata',
                        ['key', 'value'])
    if 'ix_test_key_value' not in test_metad_indx_names:
        op.create_index('ix_test_key_value', 'test_metadata', ['key', 'value'])
    # Add compound index on test_id and status and start_time, these are common
    # graph query patterns
    if 'ix_test_id_status' not in test_run_indx_names:
        op.create_index('ix_test_id_status', 'test_runs',
                        ['test_id', 'status'], mysql_length={'status': 30})
    if 'ix_test_id_start_time' not in test_run_indx_names:
        op.create_index('ix_test_id_start_time', 'test_runs', ['test_id',
                                                               'start_time'])


def downgrade():
    op.drop_index('ix_test_start_time', 'test_runs')
    op.drop_index('ix_test_stop_time', 'test_runs')
    op.drop_index('ix_run_at', 'runs')
    op.drop_index('ix_run_key_value', 'run_metadata')
    op.drop_index('ix_test_run_key_value', 'test_run_metadata')
    op.drop_index('ix_test_key_value', 'test_metadata')
    op.drop_index('ix_test_id_status', 'test_runs')
    op.drop_index('ix_test_id_start_time', 'test_runs')



# Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import datetime
import sys

from dateutil import parser as date_parser
from oslo_config import cfg
from oslo_db import options
from pbr import version
from stevedore import enabled

from subunit2sql.db import api
from subunit2sql import exceptions
from subunit2sql import read_subunit as subunit

CONF = cfg.CONF
CONF.import_opt('verbose', 'subunit2sql.db.api')

SHELL_OPTS = [
    cfg.MultiStrOpt('subunit_files', positional=True,
                    help='list of subunit files to put into the database'),
    cfg.DictOpt('run_meta', short='r', default=None,
                help='Dict of metadata about the run(s)'),
    cfg.StrOpt('artifacts', short='a', default=None,
               help='Location of run artifacts'),
    cfg.BoolOpt('store_attachments', short='s', default=False,
                help='Store attachments from subunit streams in the DB'),
    cfg.StrOpt('run_id', short='i', default=None,
               help='Run id to use for the specified subunit stream, can only'
                    ' be used if a single stream is provided'),
    cfg.StrOpt('attr_regex', default='\[(.*)\]',
               help='The regex to use to extract the comma separated list of '
                    'test attributes from the test_id'),
    cfg.StrOpt('test_attr_prefix', short='p', default=None,
               help='An optional prefix to identify global test attrs '
                    'and treat it as test metadata instead of test_run '
                    'metadata'),
    cfg.BoolOpt('remove_test_attr_prefix', short='x', default=False,
                help='When True, the prefix configured in "test_attr_prefix", '
                     'if any, is removed from the metadata before it\'s '
                     'added to the test metadata'),
    cfg.StrOpt('run_at', default=None,
               help="The optional datetime string for the run was started, "
                    "If one isn't provided the date and time of when "
                    "subunit2sql is called will be used"),
    cfg.BoolOpt('use_run_wall_time', default=False, short='w',
                help="When True the wall time of a run will be used for the "
                     "run_time column in the runs table. By default the sum of"
                     " the test executions are used instead."),
    cfg.StrOpt('non_subunit_name', default=None,
               help='Allows non-subunit content and stores it under this'
               ' name'),
]

_version_ = version.VersionInfo('subunit2sql').version_string_with_vcs()


def cli_opts():
    for opt in SHELL_OPTS:
        CONF.register_cli_opt(opt)


def list_opts():
    """Return a list of oslo.config options available.

    The purpose of this is to allow tools like the Oslo sample config file
    generator to discover the options exposed to users.
    """
    return [('DEFAULT', copy.deepcopy(SHELL_OPTS))]


def parse_args(argv, default_config_files=None):
    cfg.CONF.register_cli_opts(options.database_opts, group='database')
    cfg.CONF(argv[1:], project='subunit2sql', version=_version_,
             default_config_files=default_config_files)


def running_avg(test, values, result):
    count = test.success
    avg_prev = test.run_time
    curr_runtime = subunit.get_duration(result['start_time'],
                                        result['end_time'])
    if isinstance(avg_prev, float):
        # Using a smoothed moving avg to limit the affect of a single outlier
        new_avg = ((count * avg_prev) + curr_runtime) / (count + 1)
        values['run_time'] = new_avg
    else:
        values['run_time'] = curr_runtime
    return values


def increment_counts(test, results):
    test_values = {'run_count': test.run_count + 1}
    status = results.get('status')
    if status == 'success':
        test_values['success'] = test.success + 1
        test_values = running_avg(test, test_values, results)
    elif status == 'fail':
        test_values['failure'] = test.failure + 1
    elif status == 'skip':
        test_values = {}
    else:
        msg = "Unknown test status %s" % status
        raise exceptions.UnknownStatus(msg)
    return test_values


def get_run_totals(results):
    success = len([x for x in results if results[x]['status'] == 'success'])
    fails = len([x for x in results if results[x]['status'] == 'fail'])
    skips = len([x for x in results if results[x]['status'] == 'skip'])
    totals = {
        'success': success,
        'fails': fails,
        'skips': skips,
    }
    return totals


def _get_test_attrs_list(attrs):
    if attrs:
        attr_list = attrs.split(',')
        test_attrs_list = [attr for attr in attr_list if attr.startswith(
            CONF.test_attr_prefix)]
        return test_attrs_list
    else:
        return None


def _override_conf(value, value_name):
    if (not value) and hasattr(CONF, value_name):
        return CONF[value_name]
    else:
        return value


def process_results(results, run_at=None, artifacts=None, run_id=None,
                    run_meta=None, test_attr_prefix=None):
    """Insert converted subunit data into the database.

    Allows for run-specific information to be passed in via kwargs,
    checks CONF if no run-specific information is supplied.

    :param results: subunit stream to be inserted
    :param run_at: Optional time at which the run was started.
    :param artifacts: Link to any artifacts from the test run.
    :param run_id: The run id for the new run. Must be unique.
    :param run_meta: Metadata corresponding to the new run.
    :param test_attr_prefix: Optional test attribute prefix.
    """
    run_at = _override_conf(run_at, 'run_at')
    artifacts = _override_conf(artifacts, 'artifacts')
    run_id = _override_conf(run_id, 'run_id')
    run_meta = _override_conf(run_meta, 'run_meta')
    test_attr_prefix = _override_conf(test_attr_prefix, 'test_attr_prefix')

    if run_at:
        if not isinstance(run_at, datetime.datetime):
            run_at = date_parser.parse(run_at)
    else:
        run_at = None
    session = api.get_session()
    run_time = results.pop('run_time')
    totals = get_run_totals(results)
    db_run = api.create_run(totals['skips'], totals['fails'],
                            totals['success'], run_time, artifacts,
                            id=run_id, run_at=run_at, session=session)
    if run_meta:
        api.add_run_metadata(run_meta, db_run.id, session)
    for test in results:
        db_test = api.get_test_by_test_id(test, session)
        if not db_test:
            if results[test]['status'] == 'success':
                success = 1
                fails = 0
            elif results[test]['status'] == 'fail':
                fails = 1
                success = 0
            else:
                fails = 0
                success = 0
            run_time = subunit.get_duration(results[test]['start_time'],
                                            results[test]['end_time'])
            db_test = api.create_test(test, (success + fails), success,
                                      fails, run_time,
                                      session)
        else:
            test_values = increment_counts(db_test, results[test])
            # If skipped nothing to update
            if test_values:
                api.update_test(test_values, db_test.id, session)
        test_run = api.create_test_run(db_test.id, db_run.id,
                                       results[test]['status'],
                                       results[test]['start_time'],
                                       results[test]['end_time'],
                                       session)
        if results[test]['metadata']:
            if test_attr_prefix:
                attrs = results[test]['metadata'].get('attrs')
                test_attr_list = _get_test_attrs_list(attrs)
                test_metadata = api.get_test_metadata(db_test.id, session)
                test_metadata = [(meta.key, meta.value) for meta in
                                 test_metadata]
                if test_attr_list:
                    for attr in test_attr_list:
                        if CONF.remove_test_attr_prefix:
                            normalized_attr = attr[len(
                                CONF.test_attr_prefix):]
                        else:
                            normalized_attr = attr
                        if ('attr', normalized_attr) not in test_metadata:
                            test_meta_dict = {'attr': normalized_attr}
                            api.add_test_metadata(test_meta_dict, db_test.id,
                                                  session=session)
            api.add_test_run_metadata(results[test]['metadata'], test_run.id,
                                      session)
        if results[test]['attachments']:
            api.add_test_run_attachments(results[test]['attachments'],
                                         test_run.id, session)
    session.close()
    return db_run


def get_extensions():
    def check_enabled(ext):
        return ext.plugin.enabled()
    return enabled.EnabledExtensionManager('subunit2sql.target',
                                           check_func=check_enabled)


def get_targets(extensions):
    try:
        targets = list(extensions.map(lambda ext: ext.plugin()))
    except RuntimeError:
        targets = []
    return targets


def main():
    cli_opts()

    extensions = get_extensions()
    parse_args(sys.argv)
    targets = get_targets(extensions)
    if CONF.subunit_files:
        if len(CONF.subunit_files) > 1 and CONF.run_id:
            print("You can not specify a run id for adding more than 1 stream")
            return 3
        streams = [subunit.ReadSubunit(open(s, 'r'),
                                       attachments=CONF.store_attachments,
                                       attr_regex=CONF.attr_regex,
                                       targets=targets,
                                       use_wall_time=CONF.use_run_wall_time,
                                       non_subunit_name=CONF.non_subunit_name)
                   for s in CONF.subunit_files]
    else:
        streams = [subunit.ReadSubunit(sys.stdin,
                                       attachments=CONF.store_attachments,
                                       attr_regex=CONF.attr_regex,
                                       targets=targets,
                                       use_wall_time=CONF.use_run_wall_time,
                                       non_subunit_name=CONF.non_subunit_name)]
    for stream in streams:
        process_results(stream.get_results())


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

import matplotlib
import matplotlib.pyplot as plt
from oslo_config import cfg
import pandas as pd

from subunit2sql.db import api

CONF = cfg.CONF
matplotlib.style.use('ggplot')


def set_cli_opts(parser):
    parser.add_argument('metadata_key',
                        help="The run_metadata key to group the runs by")
    parser.add_argument('--filter_list', '-f',
                        help='A comma seperated list of values to use')


def generate_series():
    session = api.get_session()
    if CONF.command.filter_list:
        filter_list = CONF.command.filter_list.split(',')
    else:
        filter_list = None
    if CONF.start_date:
        start_date = datetime.datetime.strptime(CONF.start_date, '%Y-%m-%d')
    else:
        start_date = None
    if CONF.stop_date:
        stop_date = datetime.datetime.strptime(CONF.stop_date, '%Y-%m-%d')
    else:
        stop_date = None
    run_times = api.get_run_times_grouped_by_run_metadata_key(
        CONF.command.metadata_key, start_date=start_date,
        stop_date=stop_date, session=session)
    if not filter_list:
        df = pd.DataFrame(dict(
            [(k, pd.Series(v)) for k, v in run_times.iteritems()]))
    else:
        df = pd.DataFrame(dict(
            [(k, pd.Series(v)) for k, v in run_times.iteritems()
             if k in filter_list]))
    if not CONF.title:
        title = "Run aggregate run time grouped by metadata"
    else:
        title = CONF.title
    # NOTE(mtreinish): Decrease label font size for the worst case where we
    # have tons of groups
    matplotlib.rcParams['xtick.labelsize'] = '3'
    plt.figure()
    plt.title(title)
    df.plot(kind='box', rot=90)
    plt.ylabel('Time (sec.)')
    plt.tight_layout()
    plt.savefig(CONF.output, dpi=CONF.dpi)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_config import cfg

CONF = cfg.CONF


def filter_dates(ts):
    if CONF.start_date:
        start_date = datetime.datetime.strptime(CONF.start_date, '%Y-%m-%d')
        if CONF.stop_date:
            stop_date = datetime.datetime.strptime(CONF.stop_date, '%Y-%m-%d')
            ts = ts.ix[start_date:stop_date]
        else:
            ts = ts.ix[start_date:]

    elif CONF.stop_date:
        stop_date = datetime.datetime.strptime(CONF.stop_date, '%Y-%m-%d')
        ts = ts.ix[:stop_date]
    return ts
